"""Tool to search keybinds that match given search criteria."""
import argparse
import subprocess
import sys
import tempfile
from collections import deque
from dataclasses import dataclass
from typing import (
    Any,
    ClassVar,
    Deque,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    cast,
)

from ..errors import SXHKDParserError
from ..metadata import SectionTreeNode
from ..parser import Keybind
from ..util import read_sxhkdrc
from .common import (
    BASE_PARSER,
    IGNORE_HOTKEY_ERRORS,
    ReplaceStrEvaluator,
    add_repl_str_options,
    format_error_msg,
    get_command_name,
    process_args,
)

__all__ = ["main"]


@dataclass
class KeybindMatcherContext:
    replace_str: ReplaceStrEvaluator


class Expression:
    @classmethod
    def from_args(cls, args: Deque[str]) -> "Expression":
        return OrExpression.from_args(args)

    # path from root to here (last element is current section)
    # empty path means root
    # RETURN: permutation indices that match
    def match(
        self, keybind: Keybind, path: List[str], ctx: KeybindMatcherContext
    ) -> Set[int]:
        raise NotImplementedError

    def print(self, level: int = 0) -> None:
        raise NotImplementedError


RESERVED_ARGS = set()


@dataclass
class OrExpression(Expression):
    lhs: Expression
    rhs: Expression
    RESERVED_ARGS.add("-o")

    @classmethod
    def from_args(cls, args: Deque[str]) -> Expression:
        assert args[0] not in ("-o", "-a"), args
        maybe_lhs = AndExpression.from_args(args)
        # Backtrack in case this is not an OR.
        if not args:
            return maybe_lhs

        out_expr: Expression
        if args[0] == "-o":
            args.popleft()
            rhs = cls.from_args(args)
            out_expr = cls(maybe_lhs, rhs)
            return out_expr
        else:
            return AndExpression.from_args(args, lhs=maybe_lhs)

    def match(
        self, keybind: Keybind, path: List[str], ctx: KeybindMatcherContext
    ) -> Set[int]:
        # Short-circuit.
        lhs = self.lhs.match(keybind, path, ctx)
        if lhs == set(range(len(keybind.hotkey.permutations))):
            return lhs
        return lhs | self.rhs.match(keybind, path, ctx)

    def print(self, level: int = 0) -> None:
        print(f"{' ' * level}or-lhs:")
        self.lhs.print(level + 1)
        print(f"{' ' * level}or-rhs:")
        self.rhs.print(level + 1)


@dataclass
class AndExpression(Expression):
    lhs: Expression
    rhs: Expression
    RESERVED_ARGS.add("-a")

    @classmethod
    def from_args(
        cls, args: Deque[str], lhs: Optional[Expression] = None
    ) -> Expression:
        assert args[0] != "-o"
        if lhs is None:
            lhs = NotExpression.from_args(args)
            # Backtrack in case this is not an AND.
            if not args:
                return lhs
        # Allow ANDing via juxtaposition
        if args[0] == "-a":
            args.popleft()
        # `lhs` should be LHS of calling function (parse_or).  AND binds tighter
        # than OR, and need to take into account surrounding parens.  Not returning
        # early would cause parse errors in the higher-precedence operators or in
        # the predicates.
        elif args[0] in ("-o", ")"):
            return lhs
        rhs = cls.from_args(args)
        out_expr = cls(lhs, rhs)
        return out_expr

    def match(
        self, keybind: Keybind, path: List[str], ctx: KeybindMatcherContext
    ) -> Set[int]:
        # Short-circuit.
        lhs = self.lhs.match(keybind, path, ctx)
        if not lhs:
            return lhs
        return lhs & self.rhs.match(keybind, path, ctx)

    def print(self, level: int = 0) -> None:
        print(f"{' ' * level}and-lhs:")
        self.lhs.print(level + 1)
        print(f"{' ' * level}and-rhs:")
        self.rhs.print(level + 1)


@dataclass
class NotExpression(Expression):
    subexpr: Expression
    RESERVED_ARGS.add("!")

    @classmethod
    def from_args(cls, args: Deque[str]) -> Expression:
        if args[0] == "!":
            args.popleft()
            subexpr = cls.from_args(args)
            return cls(subexpr)
        else:
            return _ParenExpression.from_args(args)

    def match(
        self, keybind: Keybind, path: List[str], ctx: KeybindMatcherContext
    ) -> Set[int]:
        # return not self.subexpr.match(keybind, path, ctx)
        return set(
            range(len(keybind.hotkey.permutations))
        ) - self.subexpr.match(keybind, path, ctx)

    def print(self, level: int = 0) -> None:
        print(f"{' ' * level}not:")
        self.subexpr.print(level + 1)


# RESERVED_ARGS.add("(")
# RESERVED_ARGS.add(")")


@dataclass
class _ParenExpression(Expression):
    subexpr: Expression
    RESERVED_ARGS.add("(")
    RESERVED_ARGS.add(")")

    def __init__(self, *args: Any, **kwargs: Any):
        raise RuntimeError("This should never have been instantiated")

    @classmethod
    def from_args(cls, args: Deque[str]) -> Expression:
        if args[0] == "(":
            args.popleft()
            subexpr = Expression.from_args(args)
            maybe_rparen = args.popleft()
            if maybe_rparen == ")":
                # return ParenExpression(subexpr)
                return subexpr
            else:
                raise ValueError(f"expected ')' but got {maybe_rparen!r}")
        else:
            return PredicateExpression.from_args(args)

    def match(
        self, keybind: Keybind, path: List[str], ctx: KeybindMatcherContext
    ) -> Set[int]:
        return self.subexpr.match(keybind, path, ctx)

    def print(self, level: int = 0) -> None:
        print(f"{' ' * level}paren:")
        self.subexpr.print(level + 1)


class PredicateExpression(Expression):
    _PARSERS: ClassVar[Dict[str, Type["PredicateExpression"]]] = {}
    RESERVED: ClassVar[Sequence[str]] = ()

    def __init_subclass__(cls, **kwargs: Any):
        for reserved_arg in getattr(cls, "RESERVED", []):
            cls._PARSERS[reserved_arg] = cls

    @classmethod
    def from_args(cls, args: Deque[str]) -> Expression:
        val = args.popleft()
        try:
            pred_expr_cls = cls._PARSERS[val]
        except KeyError as e:
            raise ValueError(f"invalid syntax item: {val!r}") from e
        else:
            return pred_expr_cls.from_args(args)


@dataclass
class CommandPredicate(PredicateExpression):
    args: List[str]
    RESERVED = ["-cmd"]
    RESERVED_ARGS.update(RESERVED)

    def __init__(self, args: List[str]):
        if not args:
            raise ValueError("empty command")
        self.args = args

    @classmethod
    def from_args(cls, args: Deque[str]) -> Expression:
        cmdargs = []
        while args and args[0] != ";":
            cmdargs.append(args.popleft())
        if not args:
            raise ValueError(
                f"unterminated command ({'|'.join(cls.RESERVED)}) predicate"
            )
        args.popleft()
        return cls(cmdargs)

    def match(
        self, keybind: Keybind, path: List[str], ctx: KeybindMatcherContext
    ) -> Set[int]:
        matches = set()
        for i, (hk, cmd) in enumerate(
            zip(keybind.hotkey.permutations, keybind.command.permutations)
        ):
            with tempfile.NamedTemporaryFile("w+") as f:
                f.write(cmd)
                f.flush()
                if not any(
                    ctx.replace_str.hotkey in arg
                    or ctx.replace_str.command in arg
                    for arg in self.args
                ):
                    cmdline = self.args + [f.name]
                else:
                    norm_str = str(hk)
                    cmdline = ctx.replace_str.eval(
                        self.args, hotkey=norm_str, cmd_file=f.name
                    )
                try:
                    result = subprocess.run(
                        cmdline,
                        capture_output=True,
                        text=True,
                        check=True,
                    )
                except subprocess.CalledProcessError as e:
                    print(e.stderr, end="", file=sys.stderr)
                    continue
                except FileNotFoundError as e:
                    print(e, file=sys.stderr)
                    continue
                else:
                    print(result.stderr, end="", file=sys.stderr)
                    matches.add(i)
        return matches

    def print(self, level: int = 0) -> None:
        print(f"{' ' * level}command: {self.args}")


@dataclass
class MetadataPredicate(PredicateExpression):
    name: str
    operator: Optional[str]
    value: Optional[str]
    OPERATORS: ClassVar[Sequence[str]] = ("=", "!=")
    RESERVED = ["-has"]
    RESERVED_ARGS.update(RESERVED)

    def __init__(
        self,
        name: str,
        operator: Optional[str] = None,
        value: Optional[str] = None,
    ):
        if value is not None:
            assert operator is not None
        self.name = name
        if value is not None and operator not in MetadataPredicate.OPERATORS:
            raise ValueError(
                f"invalid operator {operator!r}: must be one of {MetadataPredicate.OPERATORS}"
            )
        self.operator = operator
        self.value = value

    @classmethod
    def from_args(cls, args: Deque[str]) -> Expression:
        name = args.popleft()
        # -has should only consume further arguments if the next one cannot
        # possibly be interpreted as another predicate or as a logical
        # operator.
        if not args or (args[0].startswith("-") or args[0] in ("!", "(", ")")):
            return cls(name)
        operator = args.popleft()
        value = args.popleft()
        return cls(name, operator, value)

    def match(
        self, keybind: Keybind, path: List[str], ctx: KeybindMatcherContext
    ) -> Set[int]:
        perms = set(range(len(keybind.hotkey.permutations)))
        # Existence (-has NAME)
        if self.value is None:
            if self.name in keybind.metadata:
                return perms
            else:
                return set()

        # Others (-has NAME(=|!=)VALUE)
        if self.operator == "=":
            if keybind.metadata.get(self.name) == self.value:
                return perms
            else:
                return set()
        elif self.operator == "!=":
            if keybind.metadata.get(self.name) != self.value:
                return perms
            else:
                return set()
        else:
            raise NotImplementedError

    def print(self, level: int = 0) -> None:
        if self.value:
            print(
                f"{' ' * level}metadata-value: {self.name!r} {self.operator} {self.value!r}"
            )
        else:
            print(f"{' ' * level}metadata-exists: {self.name!r}")


def parse_path(raw_path: str) -> Tuple[List[str], bool]:
    assert raw_path
    if raw_path[0] == "/":
        absolute = True
        raw_path = raw_path[1:]
    else:
        absolute = False

    path = []
    component = ""
    do_escape = False
    for c in raw_path:
        if do_escape:
            if c == "/":
                component += c
            else:
                component += "\\" + c
            do_escape = False
        else:
            if c == "\\":
                do_escape = True
            elif c == "/":
                if component:
                    path.append(component)
                component = ""
            else:
                component += c
    if do_escape:
        component += "\\"
    if component:
        path.append(component)
    return (path, absolute)


@dataclass
class SectionChildPredicate(PredicateExpression):
    path: List[str]
    absolute: bool
    RESERVED = ["-in", "-child-of"]
    RESERVED_ARGS.update(RESERVED)

    @classmethod
    def from_args(cls, args: Deque[str]) -> Expression:
        path, absolute = parse_path(args.popleft())
        return cls(path, absolute=absolute)

    def match(
        self, keybind: Keybind, path: List[str], ctx: KeybindMatcherContext
    ) -> Set[int]:
        perms = set(range(len(keybind.hotkey.permutations)))
        if self.absolute:
            if self.path == path:
                return perms
            else:
                return set()
        else:
            assert self.path, f"empty path {self.path}"
            if len(self.path) > len(path):
                return set()
            # Only need direct children, so just match the final path components.
            elif self.path == path[len(path) - len(self.path) :]:
                return perms
            else:
                return set()

    def print(self, level: int = 0) -> None:
        print(
            f"{' ' * level}child-of ({'absolute' if self.absolute else 'relative'}): {self.path}"
        )


@dataclass
class SectionDescendantPredicate(PredicateExpression):
    path: List[str]
    absolute: bool
    RESERVED = ["-under", "-descendant-of"]
    RESERVED_ARGS.update(RESERVED)

    @classmethod
    def from_args(cls, args: Deque[str]) -> Expression:
        path, absolute = parse_path(args.popleft())
        return cls(path, absolute=absolute)

    def match(
        self, keybind: Keybind, path: List[str], ctx: KeybindMatcherContext
    ) -> Set[int]:
        perms = set(range(len(keybind.hotkey.permutations)))
        if len(path) <= len(self.path):
            return set()
        if self.absolute:
            if self.path == path[: len(self.path)]:
                return perms
            else:
                return set()
        else:
            assert self.path, f"empty path {self.path}"
            # WARNING: very bad time-complexity (roughly O(nm) where n=len(path), m=len(self.path)).
            # Might not be so bad in practice since section nesting won't be too deep (2-3 at most), I predict.
            # TODO: maybe use smarter substring finder algorithm (but can't use Python strings---each elem is a path component).
            subpath = path
            # Need non-direct descendants, so can't match on final path components (must use strict >).
            while len(subpath) > len(self.path):
                if self.path == subpath[: len(self.path)]:
                    return perms
                subpath = subpath[1:]
            return set()

    def print(self, level: int = 0) -> None:
        print(
            f"{' ' * level}descendant-of ({'absolute' if self.absolute else 'relative'}): {self.path}"
        )


def match_keybinds(
    node: SectionTreeNode,
    expr: Optional[Expression],
    path: List[str],
    ctx: KeybindMatcherContext,
) -> Iterable[Tuple[Keybind, Set[int]]]:
    for keybind in node.keybind_children:
        if expr is None:
            matches = set(range(len(keybind.hotkey.permutations)))
        else:
            matches = expr.match(keybind, path, ctx)
        yield (keybind, matches)
    for subsection in node.children:
        yield from match_keybinds(
            subsection, expr, path + [cast(str, subsection.name)], ctx
        )


def main(argv: Optional[List[str]] = None) -> int:
    """Run the command-line tool with the given arguments, without the command name.

    Currently only returns 0 (success) and any non-SXHKDParserError exceptions
    are left unhandled.
    """
    parser = argparse.ArgumentParser(
        get_command_name(__file__),
        usage="%(prog)s [option ...] [expression]",
        description="Find hotkeys that match a search expression",
        epilog="See %(prog)s(1) for the expression syntax.",
        parents=[BASE_PARSER],
    )
    add_repl_str_options(parser)
    parser.add_argument(
        "--print-expression-tree",
        "-T",
        action="store_true",
        help="print parsed tree for the search expression and exit",
    )

    # Separate options from the search expression.
    if argv is None:
        argv = sys.argv[1:]
    argv: Deque[str] = deque(argv)
    options = []
    while argv and argv[0] not in RESERVED_ARGS:
        curr = argv.popleft()
        if curr == "--":
            break
        options.append(curr)
    raw_expr = argv

    namespace = parser.parse_args(options)
    section_handler, metadata_parser = process_args(namespace)
    expr: Optional[Expression]
    if raw_expr:
        expr = Expression.from_args(raw_expr)
    else:
        expr = None
    if namespace.print_expression_tree:
        if expr is not None:
            expr.print()
        return 0

    for bind_or_err in read_sxhkdrc(
        namespace.sxhkdrc,
        section_handler=section_handler,
        metadata_parser=metadata_parser,
        # Handle them ourselves.
        hotkey_errors=IGNORE_HOTKEY_ERRORS,
    ):
        if isinstance(bind_or_err, SXHKDParserError):
            msg = format_error_msg(bind_or_err, namespace.sxhkdrc)
            print(msg, file=sys.stderr)
            continue

    for keybind, matches in match_keybinds(
        section_handler.root,
        expr,
        path=[],
        ctx=KeybindMatcherContext(
            ReplaceStrEvaluator(
                namespace.hotkey_replace_str, namespace.command_replace_str
            )
        ),
    ):
        hk = keybind.hotkey
        for i in matches:
            print(hk.permutations[i])

    return 0



"""Tool for exporting hotkeys to various formats, including HTML and plaintext."""
from __future__ import annotations

import argparse
import sys
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Dict, Iterable, List, Optional, Set, Type

from ..errors import SXHKDParserError
from ..metadata import SectionTreeNode
from ..parser import expand_sequences
from ..util import read_sxhkdrc
from .common import (
    BASE_PARSER,
    IGNORE_HOTKEY_ERRORS,
    format_error_msg,
    get_command_name,
    print_exceptions,
    process_args,
)

__all__ = ["main"]


class PrintOption(Enum):
    SECTIONS = auto()
    KEYBINDS = auto()


class KeybindEmitter(ABC):
    def __init__(
        self,
        options: Optional[Iterable[PrintOption]] = None,
        expand_sequences: bool = False,
    ):
        self.options: Set[PrintOption]
        if options is None:
            self.options = set()
        else:
            self.options = set(options)
        self.expand_sequences = expand_sequences

    @abstractmethod
    def emit_node_header(
        self, node: SectionTreeNode, level: int, fields: List[str]
    ) -> Iterable[str]:
        raise NotImplementedError

    @abstractmethod
    def emit_node_keybinds(
        self, node: SectionTreeNode, level: int, fields: List[str]
    ) -> Iterable[str]:
        raise NotImplementedError

    def emit_node_footer(
        self, node: SectionTreeNode, level: int, fields: List[str]
    ) -> Iterable[str]:
        return iter([])

    def emit(
        self, node: SectionTreeNode, level: int, fields: List[str]
    ) -> Iterable[str]:
        if PrintOption.SECTIONS in self.options:
            yield from self.emit_node_header(node, level=level, fields=fields)
        if PrintOption.KEYBINDS in self.options:
            yield from self.emit_node_keybinds(
                node, level=level, fields=fields
            )
        for subsection in node.children:
            yield from self.emit(subsection, level=level + 1, fields=fields)
        if PrintOption.SECTIONS in self.options:
            yield from self.emit_node_footer(node, level=level, fields=fields)


HTML_BODY_ESCAPES: Dict[int, str] = str.maketrans(
    {
        "&": "&amp;",
        "<": "&lt;",
        ">": "&gt;",
    }
)
HTML_ATTR_ESCAPES: Dict[int, str] = str.maketrans(
    {
        '"': "&quot;",
        "'": "&apos;",
    }
)


def esc_body(s: str) -> str:
    return s.translate(HTML_BODY_ESCAPES)


def esc_attr(s: str) -> str:
    return s.translate(HTML_ATTR_ESCAPES)


class HTMLEmitter(KeybindEmitter):
    def _get_id_slug(self, node: SectionTreeNode) -> str:
        assert isinstance(node.name, str)
        return node.name.lower().replace(" ", "-").replace("/", "-")

    def emit_node_header(
        self, node: SectionTreeNode, level: int, fields: List[str]
    ) -> Iterable[str]:
        header_num = 1 + level
        # clamp it
        if header_num > 6:
            header_num = 6
        assert isinstance(node.name, str)
        id_slug = self._get_id_slug(node)
        yield f'{"  " * level}<div class="section" id="{esc_attr(id_slug)}">'
        yield f'{"  " * (level+1)}<h{header_num}>{esc_body(node.name)}</h{header_num}>'

    def emit_node_keybinds(
        self, node: SectionTreeNode, level: int, fields: List[str]
    ) -> Iterable[str]:
        base = "  " * (level + 1)
        yield f'{base}<table class="hotkeys">'
        for i, keybind in enumerate(node.keybind_children):
            curr_base = base + "  "
            if self.expand_sequences:
                if "description" in keybind.metadata:
                    try:
                        desc_perms = expand_sequences(
                            keybind.metadata["description"]
                        ).generate_permutations()
                    except Exception:
                        desc_perms = None
                    else:
                        assert desc_perms is not None
                        if len(desc_perms) != len(keybind.hotkey.permutations):
                            desc_perms = None
                else:
                    desc_perms = None
                for j, perm in enumerate(keybind.hotkey.permutations):
                    yield f'{curr_base}<tr class="hotkey" id="{esc_attr(self._get_id_slug(node))}-{i}-{j}">'
                    field_base = curr_base + "  "
                    for field in fields:
                        if field == "hotkey":
                            hotkey_str = str(perm)
                            yield f'{field_base}<td class="bind">{esc_body(hotkey_str)}</td>'
                        elif field == "mode":
                            yield f'{field_base}<td class="mode">{esc_body(keybind.metadata.get("mode", "normal"))}</td>'
                        elif field == "description" and desc_perms:
                            yield f'{field_base}<td class="field-{esc_attr(field)}">{esc_body(str(desc_perms[j]))}</td>'
                        else:
                            if field in keybind.metadata:
                                body = esc_body(keybind.metadata[field])
                            else:
                                body = (
                                    f"<em>No field '{esc_body(field)}'.</em>"
                                )
                            yield f'{field_base}<td class="field-{esc_attr(field)}">{body}</td>'
                    yield f"{curr_base}</tr>"
            else:
                yield f'{curr_base}<tr class="hotkey" id="{esc_attr(self._get_id_slug(node))}-{i}">'
                field_base = curr_base + "  "
                for field in fields:
                    if field == "hotkey":
                        yield f'{field_base}<td class="bind">{esc_body(keybind.hotkey.raw if isinstance(keybind.hotkey.raw, str) else " ".join(keybind.hotkey.raw))}</td>'
                    elif field == "mode":
                        yield f'{field_base}<td class="mode">{esc_body(keybind.metadata.get("mode", "normal"))}</td>'
                    else:
                        if field in keybind.metadata:
                            body = esc_body(keybind.metadata[field])
                        else:
                            body = f"<em>No field '{esc_body(field)}'.</em>"
                        yield f'{field_base}<td class="field-{esc_attr(field)}">{body}</td>'
                yield f"{curr_base}</tr>"
        yield f"{base}</table>"

    def emit_node_footer(
        self, node: SectionTreeNode, level: int, fields: List[str]
    ) -> Iterable[str]:
        yield f"{'  ' * level}</div>"


class PlaintextEmitter(KeybindEmitter):
    def emit_node_header(
        self, node: SectionTreeNode, level: int, fields: List[str]
    ) -> Iterable[str]:
        assert isinstance(node.name, str)
        yield f"{'#' * (level+1)}{node.name}"

    def emit_node_keybinds(
        self, node: SectionTreeNode, level: int, fields: List[str]
    ) -> Iterable[str]:
        for keybind in node.keybind_children:
            if self.expand_sequences:
                if "description" in keybind.metadata:
                    try:
                        desc_perms = expand_sequences(
                            keybind.metadata["description"]
                        ).generate_permutations()
                    except Exception:
                        desc_perms = None
                    else:
                        assert desc_perms is not None
                        if len(desc_perms) != len(keybind.hotkey.permutations):
                            desc_perms = None
                else:
                    desc_perms = None
                for j, perm in enumerate(keybind.hotkey.permutations):
                    line = []
                    for field in fields:
                        if field == "hotkey":
                            line.append(str(perm))
                        elif field == "mode":
                            line.append(keybind.metadata.get("mode", "normal"))
                        elif field == "description" and desc_perms:
                            line.append(str(desc_perms[j]))
                        else:
                            line.append(keybind.metadata.get(field, ""))
                    yield "\t".join(line)
            else:
                line = []
                for field in fields:
                    if field == "hotkey":
                        if isinstance(keybind.hotkey.raw, str):
                            line.append(keybind.hotkey.raw)
                        else:
                            line.append(" ".join(keybind.hotkey.raw))
                    elif field == "mode":
                        line.append(keybind.metadata.get("mode", "normal"))
                    else:
                        line.append(keybind.metadata.get(field, ""))
                yield "\t".join(line)


def main(argv: Optional[List[str]] = None) -> int:
    """Run the command-line tool with the given arguments, without the command name.

    Currently only returns 0 (success) and any non-SXHKDParserError exceptions
    are left unhandled.  SXHKDParserError instances yielded by `read_sxhkdrc`
    are simply printed to stderr.
    """
    parser = argparse.ArgumentParser(
        get_command_name(__file__),
        description="Export keybinds in various formats",
        parents=[BASE_PARSER],
    )
    parser.add_argument(
        "--format",
        "-f",
        choices=["html", "txt"],
        default="html",
        help="the format to export to (default: %(default)s)",
    )
    parser.add_argument(
        "--fields",
        "-F",
        default="hotkey,description,mode",
        help="comma-separated list of the metadata fields to include and the order in which to print them ('hotkey' isn't strictly metadata, but oh well) (default: %(default)s)",
    )
    parser.add_argument(
        "--expand",
        "-E",
        action="store_true",
        help="expand embedded sequences",
    )
    records_group = parser.add_mutually_exclusive_group()
    records_group.add_argument(
        "--records",
        "-R",
        choices=["all", "sections", "keybinds"],
        default="all",
        help="what to print (default: %(default)s)",
    )
    records_group.add_argument(
        "--sections-only",
        "-S",
        dest="records",
        action="store_const",
        const="sections",
        help="alias for --records=sections",
    )
    records_group.add_argument(
        "--keybinds-only",
        "-K",
        dest="records",
        action="store_const",
        const="keybinds",
        help="alias for --records=keybinds",
    )

    # TODO: allow configuring whether to specially process 'mode' in the metadata
    #   - maybe allow configuring what key it is?

    namespace = parser.parse_args(argv)
    namespace.fields = namespace.fields.split(",")
    section_handler, metadata_parser = process_args(namespace)

    # Exhaust the generator function to read all keybinds.
    try:
        for bind_or_err in read_sxhkdrc(
            namespace.sxhkdrc,
            section_handler=section_handler,
            metadata_parser=metadata_parser,
            hotkey_errors=IGNORE_HOTKEY_ERRORS,
        ):
            if isinstance(bind_or_err, SXHKDParserError):
                msg = format_error_msg(bind_or_err, namespace.sxhkdrc)
                print(msg, file=sys.stderr)
    except SXHKDParserError as e:
        print_exceptions(e, namespace.sxhkdrc, file=sys.stderr)
        return 1
    section_handler.root.name = "SXHKD keybinds"

    emittercls: Type[KeybindEmitter]
    if namespace.format == "html":
        emittercls = HTMLEmitter
    elif namespace.format == "txt":
        emittercls = PlaintextEmitter
    else:
        raise RuntimeError(
            f"unreachable! invalid export format {namespace.format}"
        )

    print_options: Iterable[PrintOption]
    if namespace.records == "all":
        print_options = PrintOption.__members__.values()
    else:
        print_options = [PrintOption.__members__[namespace.records.upper()]]
    emitter = emittercls(print_options, expand_sequences=namespace.expand)

    for line in emitter.emit(
        section_handler.root, level=0, fields=namespace.fields
    ):
        print(line)
    return 0



"""Program outputting the current sxhkd mode from the status fifo."""
from __future__ import annotations

import argparse
import itertools as it
import re
import subprocess
import sys
from enum import Enum
from signal import SIGUSR1, SIGUSR2, signal
from typing import Any, List, NamedTuple, Optional, Tuple

from ..errors import SXHKDParserError
from ..metadata import MetadataParser, SectionHandler
from ..parser import (
    Hotkey,
    HotkeyPermutation,
    HotkeyTree,
    HotkeyTreeChordData,
    HotkeyTreeInternalNode,
    HotkeyTreeKeysymData,
    HotkeyTreeLeafNode,
    HotkeyTreeModifierSetData,
    HotkeyTreeNode,
)
from ..util import read_sxhkdrc
from .common import (
    BASE_PARSER,
    IGNORE_HOTKEY_ERRORS,
    find_duplicates,
    find_maybe_invalid_keysyms,
    find_prefix_conflicts,
    format_error_msg,
    get_command_name,
    print_exceptions,
    process_args,
)

__all__ = ["main"]


INTERNAL_NODES = ["modifierset", "keysym"]


def read_config(
    config: str,
    section_handler: SectionHandler,
    metadata_parser: MetadataParser,
) -> Tuple[bool, HotkeyTree]:
    errored = False
    hotkey_tree = HotkeyTree(INTERNAL_NODES)
    for bind_or_err in read_sxhkdrc(
        config,
        section_handler=section_handler,
        metadata_parser=metadata_parser,
        # Handle them ourselves.
        hotkey_errors=IGNORE_HOTKEY_ERRORS,
    ):
        if isinstance(bind_or_err, SXHKDParserError):
            msg = format_error_msg(bind_or_err, config)
            print(msg, file=sys.stderr)
            errored = True
            continue

        keybind = bind_or_err
        for err in find_maybe_invalid_keysyms(keybind):
            print(format_error_msg(err, config), file=sys.stderr)
            errored = True
        hotkey_tree.merge_hotkey(keybind.hotkey)

    for err in it.chain(
        find_duplicates(hotkey_tree), find_prefix_conflicts(hotkey_tree)
    ):
        print(format_error_msg(err, config), file=sys.stderr)
        errored = True
    return (errored, hotkey_tree)


def find_matching_modsets(
    modsetdata: HotkeyTreeModifierSetData, node: HotkeyTreeInternalNode
) -> Optional[HotkeyTreeInternalNode]:
    modset_children = node.internal_children[HotkeyTreeModifierSetData]
    modsetnode = modset_children.get(modsetdata)
    if modsetnode is not None:
        return modsetnode
    for mods, child in modset_children.items():
        if mods.value < modsetdata.value:
            return find_matching_modsets(modsetdata, child)
    return None


def match_hotkey(
    perm: HotkeyPermutation, chord_index: int, curr_level: HotkeyTreeNode
) -> Optional[HotkeyTreeNode]:
    assert perm.chords, "got empty hotkey permutation"
    if chord_index >= len(perm.chords):
        assert isinstance(curr_level.data, HotkeyTreeChordData)
        # May or may not be a leaf node: `perm` might not represent a full permutation.
        return curr_level
    if isinstance(curr_level, HotkeyTreeLeafNode):
        # `perm` hasn't been completed yet but already at a leaf.
        return None
    assert isinstance(curr_level, HotkeyTreeInternalNode)

    curr = perm.chords[chord_index]
    curr_level = find_matching_modsets(
        HotkeyTreeModifierSetData(curr.modifiers), curr_level
    )
    if curr_level is None:
        return None

    curr_level = curr_level.internal_children[HotkeyTreeKeysymData].get(
        HotkeyTreeKeysymData(curr.keysym)
    )
    if curr_level is None:
        return None

    # Now find next chord
    for child in curr_level.children:
        if not isinstance(child.data, HotkeyTreeChordData):
            continue
        if curr != child.data.value:
            continue
        # Only match non-noabort nodes if perm has no noabort index.
        if perm.noabort_index is None and child.data.noabort:
            continue
        if perm.noabort_index is not None:
            # Only match noabort nodes when we are not at the index where the
            # noabort node should be.
            if chord_index != perm.noabort_index and child.data.noabort:
                continue
            if chord_index == perm.noabort_index and not child.data.noabort:
                continue
        return match_hotkey(perm, chord_index + 1, child)
    return None


class Mode(Enum):
    normal = "N"
    unknown = "U"
    inchain = "C"
    usermode = "M"


class CurrMode(NamedTuple):
    mode: Mode
    message: str

    def __str__(self) -> str:
        return f"{self.mode.value}{self.message}"


PROGNAME = get_command_name(__file__)
NOTIFY_SEND_ERR_PREFIX = [
    "notify-send",
    "-t",
    "10000",
    "-u",
    "critical",
]
NOTIFY_SEND_INFO_PREFIX = [
    "notify-send",
    "-t",
    "5000",
]


def main(argv: Optional[List[str]] = None) -> int:
    """Run the command-line tool with the given arguments, without the command name.

    Currently only returns 0 (success) and any non-SXHKDParserError exceptions
    are left unhandled.
    """
    parser = argparse.ArgumentParser(
        PROGNAME,
        description="Tail the status fifo and output the current sxhkd mode until exit.  Needs only the H*, B*, and E* fifo messages.  Send SIGUSR1 to reload config and SIGUSR2 to print the current mode.",
        parents=[BASE_PARSER],
    )
    parser.add_argument(
        "--status-fifo",
        "-s",
        default="-",
        help="the location of the status fifo ('-' for stdin) (default: %(default)s)",
    )
    parser.add_argument(
        "--mode-field",
        "-m",
        default="mode",
        help="metadata key indicating the mode that a noabort hotkey belongs to (default: %(default)s)",
    )
    parser.add_argument(
        "--notify-send-on-config-read",
        action="store_true",
        help="call `notify-send` when reading configs",
    )
    parser.add_argument(
        "--notify-send-on-config-read-error",
        action="store_true",
        help="call `notify-send` upon config read error",
    )

    namespace = parser.parse_args(argv)
    section_handler, metadata_parser = process_args(namespace)

    msg = f"Loading config file '{namespace.sxhkdrc}'"
    print(msg, file=sys.stderr)
    if namespace.notify_send_on_config_read:
        subprocess.Popen(
            NOTIFY_SEND_INFO_PREFIX
            + [
                f"{PROGNAME}: Loading config file for the first time",
                f"Reading '{namespace.sxhkdrc}'",
            ]
        )
    try:
        errored, tree = read_config(
            namespace.sxhkdrc, section_handler, metadata_parser
        )
    except Exception:
        if namespace.notify_send_on_config_read_error:
            subprocess.Popen(
                NOTIFY_SEND_ERR_PREFIX
                + [
                    f"{PROGNAME}: Fatal error upon initial config read",
                    "Quitting...",
                ]
            )
        raise
    if errored and namespace.notify_send_on_config_read_error:
        subprocess.Popen(
            NOTIFY_SEND_ERR_PREFIX
            + [
                f"{PROGNAME}: Warnings upon initial config read",
                "Read stderr for error messages",
            ]
        )

    # Actually reload the config at a set point in the loop to avoid data
    # becoming stale inside the rest of the loop body.
    # It's delayed, but it's is safe since, for us, the keybinds are
    # essentially unchanged until the first time we get a fifo message after
    # setting the flag.
    reload_config = False

    def handle_sigusr1(*_: Any) -> None:
        nonlocal reload_config
        print(
            f"Arranging to reload config file '{namespace.sxhkdrc}'",
            file=sys.stderr,
        )
        if namespace.notify_send_on_config_read:
            subprocess.Popen(
                NOTIFY_SEND_INFO_PREFIX
                + [
                    f"{PROGNAME}: Preparing to reload config file",
                    f"Will reload '{namespace.sxhkdrc}' just in time",
                ]
            )
        reload_config = True

    signal(SIGUSR1, handle_sigusr1)

    msg_re = re.compile(r"^(?P<type>[A-Z])(?P<msg>.*)$")
    try:
        if namespace.status_fifo == "-":
            status_fifo = sys.stdin
            do_close = False
        else:
            status_fifo = open(namespace.status_fifo)
            do_close = True

        prev_hotkey_str: Optional[str] = None
        curr_mode = CurrMode(Mode.normal, "normal")

        def handle_sigusr2(*_: Any) -> None:
            print(curr_mode, flush=True)

        signal(SIGUSR2, handle_sigusr2)

        print(curr_mode, flush=True)
        for line in status_fifo:
            if reload_config:
                reload_config = False
                print(
                    f"Reloading config file '{namespace.sxhkdrc}'",
                    file=sys.stderr,
                )
                if namespace.notify_send_on_config_read:
                    subprocess.Popen(
                        NOTIFY_SEND_INFO_PREFIX
                        + [
                            f"{PROGNAME}: Reloading config file",
                            f"Reading '{namespace.sxhkdrc}'",
                        ]
                    )
                new_section_handler = section_handler.clone_config()
                try:
                    errored, new_tree = read_config(
                        namespace.sxhkdrc, new_section_handler, metadata_parser
                    )
                except Exception as e:
                    print_exceptions(e, namespace.sxhkdrc, file=sys.stderr)
                    print(
                        "Got errors while reloading config: using old keybinds...",
                        file=sys.stderr,
                    )
                    if namespace.notify_send_on_config_read_error:
                        subprocess.Popen(
                            NOTIFY_SEND_ERR_PREFIX
                            + [
                                f"{PROGNAME}: Fatal config reload error",
                                "Rolling back to old keybinds...",
                            ]
                        )
                else:
                    tree = new_tree
                    section_handler = new_section_handler
                    if errored and namespace.notify_send_on_config_read_error:
                        subprocess.Popen(
                            NOTIFY_SEND_ERR_PREFIX
                            + [
                                f"{PROGNAME}: Warnings upon config reload",
                                "Read stderr for error messages",
                            ]
                        )

            m = msg_re.match(line)
            if not m:
                continue
            type_ = m.group("type")
            msg = m.group("msg")
            if type_ == "B" and msg == "Begin chain":
                assert (
                    prev_hotkey_str is not None
                ), "expected to see a chord before 'BBegin chain' but got none"
                tokens = Hotkey.tokenize(prev_hotkey_str)
                perm = Hotkey.parse_hotkey_permutation(tokens)
                # sxhkd never seems to print chains with ':' to its status fifo.
                # Also, it doesn't seem to print all hotkeys in a mode.
                # Therefore this shouldn't be a problem anyway.
                if perm.noabort_index is not None:
                    continue

                # Try to match a mode first.
                perm.noabort_index = len(perm.chords) - 1
                node = match_hotkey(perm, 0, tree.root)
                perm.noabort_index = None
                if node is None:
                    node = match_hotkey(perm, 0, tree.root)

                if node is not None:
                    assert isinstance(
                        node.data, HotkeyTreeChordData
                    ), f"got non-chord output ({node.data!r}) from match"
                    if node.data.noabort:
                        assert isinstance(node, HotkeyTreeInternalNode)
                        perm_ends = node.find_permutation_ends()
                        assert perm_ends
                        hotkey = perm_ends[0].hotkey
                        keybind = hotkey.keybind
                        assert keybind is not None
                        nodemode = keybind.metadata.get(namespace.mode_field)
                        if nodemode is None:
                            curr_mode = CurrMode(Mode.unknown, str(perm))
                        else:
                            curr_mode = CurrMode(Mode.usermode, nodemode)
                    else:
                        curr_mode = CurrMode(Mode.inchain, "in-chain")
                    print(curr_mode)
            elif type_ == "E" and msg == "End chain":
                # Timeout immediately has 'EEnd chain' after it.
                curr_mode = CurrMode(Mode.normal, "normal")
                print(curr_mode)
            elif type_ == "H":
                prev_hotkey_str = msg
            sys.stdout.flush()
    finally:
        if do_close:
            status_fifo.close()

    return 0



"""Tool to run commands on the command text of keybinds."""
from __future__ import annotations

import argparse
import os
import re
import subprocess
import sys
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from enum import Enum, auto
from typing import (
    IO,
    ClassVar,
    Container,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from ..errors import SXHKDParserError
from ..parser import Hotkey, Keybind
from ..util import read_sxhkdrc
from .common import (
    BASE_PARSER,
    IGNORE_HOTKEY_ERRORS,
    ReplaceStrEvaluator,
    add_repl_str_options,
    format_error_msg,
    get_command_name,
    process_args,
)

__all__ = ["main"]

# Same exit codes as GNU xargs (which is based on those of POSIX).
CODE_1_125 = 123
CODE_255 = 124
CODE_SIGNAL = 125
CODE_CANNOT_RUN = 126
CODE_NOT_FOUND = 127
CODE_OTHER = 1


KeybindDict = Dict[str, Tuple[Keybind, str, bool]]


def get_perms_to_exec(file: IO[str], keybinds: KeybindDict) -> Iterable[str]:
    for line in file:
        if not line.strip():
            continue
        # Normalise
        tokens = Hotkey.tokenize(line)
        norm_str = str(Hotkey.parse_hotkey_permutation(tokens))
        if norm_str in keybinds:
            yield norm_str


class LinterFormatParseMode(Enum):
    NORMAL = auto()
    NEED_LBRACE = auto()
    NEED_CHAR = auto()
    NEED_CHAR_OR_RBRACE = auto()


@dataclass
class LinterFormatField:
    field: str
    DEFINED: ClassVar[Container[str]] = {
        "file",
        "line",
        "column",
        "type",
        "message",
    }

    def __init__(self, field: str):
        self.field = field

    def validate(self) -> None:
        if self.field not in self.DEFINED:
            raise ValueError(f"undefined field {self.field!r}")


def parse_linter_format(format: str) -> List[Union[str, LinterFormatField]]:
    spans: List[Union[str, LinterFormatField]] = [""]
    mode = LinterFormatParseMode.NORMAL
    for c in format:
        if mode == LinterFormatParseMode.NORMAL:
            assert isinstance(
                spans[-1], str
            ), f"non-str latest span in mode {mode}"

            if c == "%":
                mode = LinterFormatParseMode.NEED_LBRACE
            else:
                spans[-1] += c
        elif mode == LinterFormatParseMode.NEED_LBRACE:
            assert isinstance(
                spans[-1], str
            ), f"non-str latest span in mode {mode}"

            # Allow escaping '%' chars.
            if c == "%":
                spans[-1] += c
                mode = LinterFormatParseMode.NORMAL
            elif c == "{":
                if spans[-1] == "":
                    spans.pop()
                spans.append(LinterFormatField(""))
                mode = LinterFormatParseMode.NEED_CHAR
            else:
                raise ValueError(c)
        elif mode == LinterFormatParseMode.NEED_CHAR:
            assert c not in (
                "{",
                "}",
                "%",
            ), f"special char in field name {c!r}"
            assert isinstance(
                spans[-1], LinterFormatField
            ), f"non-field-span latest span in mode {mode}"

            spans[-1].field += c
            mode = LinterFormatParseMode.NEED_CHAR_OR_RBRACE
        elif mode == LinterFormatParseMode.NEED_CHAR_OR_RBRACE:
            assert c not in ("{", "%"), f"special char in field name {c!r}"
            assert isinstance(
                spans[-1], LinterFormatField
            ), f"non-field-span latest span in mode {mode}"
            assert spans[-1].field, f"empty current field name {spans[-1]!r}"

            if c == "}":
                spans[-1].validate()
                spans.append("")
                mode = LinterFormatParseMode.NORMAL
            else:
                spans[-1].field += c
    if isinstance(spans[-1], str) and spans[-1] == "":
        spans.pop()
    return spans


def linter_format_to_regex(
    spans: List[Union[str, LinterFormatField]], filename: str
) -> re.Pattern[str]:
    patt = ""
    for span in spans:
        if isinstance(span, str):
            patt += re.escape(span)
        else:
            if span.field == "file":
                inner_re = re.escape(filename)
            elif span.field in ("line", "column"):
                inner_re = r"\d+"
            else:
                inner_re = ".+"
            patt += f"(?P<{span.field}>{inner_re})"
    return re.compile(patt)


def print_linter_output(
    keybind: Keybind,
    namespace: argparse.Namespace,
    filename: str,
    output: Iterable[str],
) -> None:
    regex = linter_format_to_regex(namespace.linter_format, filename)
    for line in output:
        m = regex.match(line)
        if m:
            msg = m.group("message")
            if "type" in regex.groupindex and m.group("type"):
                print(
                    f"{namespace.sxhkdrc}:{keybind.command.line}: {m.group('type')}: {msg}"
                )
            else:
                print(f"{namespace.sxhkdrc}:{keybind.command.line}:{msg}")


def wait_on_proc(proc: subprocess.Popen[str]) -> Tuple[str, Optional[int]]:
    _, errs = proc.communicate()
    code = None
    if proc.returncode != 0:
        if 1 <= proc.returncode <= 125:
            code = CODE_1_125
        elif proc.returncode < 0:
            code = CODE_SIGNAL
        else:
            code = CODE_CANNOT_RUN
        code -= 64
    return (errs, code)


def main(argv: Optional[List[str]] = None) -> int:
    """Run the command-line tool with the given arguments, without the command name.

    Exit codes correspond to that of xargs, subtracting 64 for errors during
    commands run due to --exec/-e.

    However, if `Exception` is raised during the intrinsic command invocations
    (from argv), exits early and exit code 1 is returned.  `Exception` during
    --exec/-e commands also exit early and then return the exit code of the
    instrinsic invocation errors + 1 (i.e., exit code 2).

    ENOENT during intrinsic invocations exit early.  Various other software
    errors return exit code 1.
    """
    parser = argparse.ArgumentParser(
        get_command_name(__file__),
        description="Run commands on sxhkdrc keybinds",
        parents=[BASE_PARSER],
    )
    add_repl_str_options(parser)
    parser.add_argument(
        "--mode",
        "-m",
        # "edit [commands]"
        choices=["edit", "filter", "linter"],
        default="edit",
        help="set mode (default: %(default)s)",
    )
    parser.add_argument(
        "--exclude-synchronous-marker",
        "-s",
        action="store_true",
        help="exclude ';' prefix from output for keybind commands that indicate synchronous",
    )
    parser.add_argument(
        "--linter-format",
        "-f",
        default="%{file}:%{line}:%{column}:%{message}",
        type=parse_linter_format,
        help="set linter error message format (default: %(default)s)",
    )
    parser.add_argument(
        "--exec",
        "-e",
        action="store_true",
        help="execute keybind command text after modification by the command from argv",
    )
    parser.add_argument(
        "--delay",
        "-d",
        default=0,
        type=float,
        help="wait DELAY seconds between each --exec command invocation (default: %(default)s)",
    )
    parser.add_argument(
        "command",
        nargs=argparse.REMAINDER,
    )
    namespace = parser.parse_args(argv)
    if namespace.command and namespace.command[0] == "--":
        namespace.command = namespace.command[1:]
    section_handler, metadata_parser = process_args(namespace)
    if namespace.mode != "edit":
        if namespace.exec:
            print(
                f"cannot use --mode={namespace.mode} with --exec",
                file=sys.stderr,
            )
            return 1

    keybinds: KeybindDict = {}
    try:
        for bind_or_err in read_sxhkdrc(
            namespace.sxhkdrc,
            section_handler=section_handler,
            metadata_parser=metadata_parser,
            # Handle them ourselves.
            hotkey_errors=IGNORE_HOTKEY_ERRORS,
        ):
            if isinstance(bind_or_err, SXHKDParserError):
                msg = format_error_msg(bind_or_err, namespace.sxhkdrc)
                print(msg, file=sys.stderr)
                continue

            for hkperm, cmdperm in zip(
                bind_or_err.hotkey.permutations,
                bind_or_err.command.permutations,
            ):
                norm_str = str(hkperm)
                keybinds[norm_str] = (
                    bind_or_err,
                    cmdperm,
                    bind_or_err.command.synchronous,
                )
    except SXHKDParserError as e:
        msg = format_error_msg(e, namespace.sxhkdrc)
        print(msg, file=sys.stderr)
        return 1

    repl_evaluator = ReplaceStrEvaluator(
        namespace.hotkey_replace_str,
        namespace.command_replace_str,
    )
    sxhkd_shell = os.getenv("SXHKD_SHELL", os.getenv("SHELL"))
    if sxhkd_shell is None:
        print("no $SXHKD_SHELL or $SHELL defined", file=sys.stderr)
        return 1
    if sxhkd_shell == "":
        print("empty $SXHKD_SHELL and $SHELL", file=sys.stderr)
        return 1

    code = 0
    procs: List[subprocess.Popen[str]] = []
    with tempfile.TemporaryDirectory() as tmpdir:
        for hk in get_perms_to_exec(sys.stdin, keybinds):
            bind, cmd, synchronous = keybinds[hk]
            with open(os.path.join(tmpdir, hk), "w") as f:
                f.write(cmd)
            if namespace.mode == "edit":
                print(hk)

            failed = False
            if namespace.command:
                # TODO: fail if hotkey repl str in cmdline but cmdline ends with `+'
                if not any(
                    namespace.hotkey_replace_str in arg
                    or namespace.command_replace_str in arg
                    for arg in namespace.command
                ):
                    cmdline = namespace.command + [f.name]
                else:
                    cmdline = repl_evaluator.eval(
                        namespace.command, hk, f.name
                    )
                try:
                    result = subprocess.run(
                        cmdline,
                        capture_output=True,
                        text=True,
                        check=True,
                    )
                except subprocess.CalledProcessError as e:
                    failed = True
                    if 1 <= e.returncode <= 125:
                        code = CODE_1_125
                    elif e.returncode == 255:
                        code = CODE_255
                    elif e.returncode < 0:
                        code = CODE_SIGNAL
                    else:
                        code = CODE_CANNOT_RUN
                    do_continue = False
                    if namespace.mode == "filter":
                        print(
                            e.stdout,
                            end="",
                        )
                    elif namespace.mode == "linter":
                        print_linter_output(
                            bind,
                            namespace,
                            filename=f.name,
                            output=e.stdout.split("\n"),
                        )
                        do_continue = True
                    print(
                        e.stderr,
                        end="",
                        file=sys.stderr,
                    )
                    # Immediately exit upon encountering 255.
                    if code == CODE_255:
                        return code
                    if do_continue:
                        continue
                except FileNotFoundError as e:
                    print(e, file=sys.stderr)
                    return CODE_NOT_FOUND
                except Exception as e:
                    print(e, file=sys.stderr)
                    return CODE_OTHER
                else:
                    if namespace.mode == "linter":
                        print_linter_output(
                            bind,
                            namespace,
                            filename=f.name,
                            output=result.stdout.split("\n"),
                        )
                        print(result.stderr, end="", file=sys.stderr)
                        continue
                    else:
                        cmd = result.stdout
                        print(result.stderr, end="", file=sys.stderr)

            assert namespace.mode != "linter"

            if namespace.exec and not failed:
                assert namespace.mode == "edit"
                cmdline = [sxhkd_shell, "-c", cmd]
                # exit codes reduced by 64 (except CODE_OTHER) to distinguish
                # which subprocess out of the intrinsic one and the --exec/-e
                # one errored.
                try:
                    newproc = subprocess.Popen(
                        cmdline,
                        text=True,
                        stdin=subprocess.DEVNULL,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.PIPE,
                        start_new_session=True,
                    )
                except FileNotFoundError as e:
                    print(e, file=sys.stderr)
                    code = CODE_NOT_FOUND - 64
                except Exception as e:
                    print(e, file=sys.stderr)
                    return CODE_OTHER + 1
                else:
                    if synchronous:
                        # Synchronous command may be long-running, so ensure it doesn't hang.
                        errs, newproc_code = wait_on_proc(newproc)
                        print(errs, end="", file=sys.stderr)
                        if newproc_code is not None:
                            code = newproc_code
                    else:
                        procs.append(newproc)
                time.sleep(namespace.delay)

            if cmd and not failed:
                # Escape braces in command.
                # NOTE: At this point, `cmd` is clean and without spans, so any
                # existing braces are literal.
                if namespace.mode == "edit":
                    cmd = cmd.replace("{", "\\{").replace("}", "\\}")

                if not namespace.exclude_synchronous_marker and synchronous:
                    cmd = f";{cmd}"

                if not cmd.endswith("\n"):
                    end = "\n"
                else:
                    end = ""
                if namespace.mode == "filter":
                    prefix = ""
                else:
                    prefix = "\t"
                print(f"{prefix}{cmd}", end=end)

    # Clean up any short-lived processes while ensuring that any long-running
    # subprocesses don't hang.
    with ThreadPoolExecutor() as executor:
        for errs, proc_code in executor.map(wait_on_proc, procs):
            print(errs, end="", file=sys.stderr)
            if proc_code is not None:
                code = proc_code

    # Just return the latest exit code.
    # POSIX xargs doesn't specify it, so this xargs-like tool can do whatever.
    return code



"""Convenience functions for using the library."""
from os import PathLike
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    TextIO,
    Union,
    cast,
)

from .errors import MissingHotkeyError, SXHKDParserError
from .metadata import MetadataParser, NullMetadataParser, SectionHandler
from .parser import Keybind

__all__ = ["read_sxhkdrc"]


def read_sxhkdrc(
    file: Union[str, PathLike[str], TextIO],
    section_handler: Optional[SectionHandler] = None,
    metadata_parser: Optional[MetadataParser] = None,
    hotkey_errors: Optional[Mapping[str, bool]] = None,
) -> Iterable[Union[SXHKDParserError, Keybind]]:
    """Parse keybinds from a given file or path, yielding a stream.

    `file` may be a filename, `os.PathLike`, or an opened file.  In the case of
    an opened file, it is not closed at the end of the function.

    A block of comments is maintained so that it can be passed to
    `MetadataParser.parse` and given to a `Keybind`.

    Upon encountering an empty line, this isolated block of comments is cleared
    so that it can't be attached to any keybinds.

    Upon encountering a comment line, the section first has its `push` method
    called on the line to define any sections if it matches.
    If a comment *doesn't* create a new section (`push` returns `False`), the
    line is added to the block of comments.
    But if a comment *does* create a new section, the block of comments is
    cleared, cutting off comments above section comments from any keybinds
    below them.

    If `section_handler` is `None`, no sections (i.e., `SectionTreeNode`
    objects) are made: they must be created manually afterwards.  Otherwise,
    the given SectionHandler instance is used to manage sections and add
    keybinds to them.

    If `metadata_parser` is `None`, all metadata comments are ignored and every
    `Keybind` instance is given an empty dict instead.
    """
    if isinstance(file, (str, PathLike)):
        f = open(file)
        close_io = True
    else:
        f = file
        # Passed in a file object.
        close_io = False
    if metadata_parser is None:
        metadata_parser = NullMetadataParser()

    hotkey: Optional[List[str]] = None
    hotkey_start_line: Optional[int] = None
    command: Optional[List[str]] = None
    command_start_line: Optional[int] = None

    comment_block_start_line: Optional[int] = None
    comment_buf: List[str] = []

    lines: List[str] = []
    line_no: int = 0
    line_block_start: int
    metadata: Dict[str, Any] = {}
    try:
        while True:
            line = f.readline()
            # readline() returns the empty string upon EOF.
            if not line:
                break
            line_no += 1
            line_block_start = line_no
            line = line.rstrip("\n")
            # sxhkd ignores empty lines.
            if not line:
                # Cut off any isolated comment blocks.
                comment_block_start_line = None
                comment_buf.clear()
                continue

            if line.startswith("#"):
                # line number so keybinds can be matched to sections later
                if section_handler is not None and section_handler.push(
                    line, line_no
                ):
                    # section handler ate it up
                    comment_block_start_line = None
                    comment_buf.clear()  # cut off any comments that could attach to a keybind
                # Metadata only comes from comments directly preceding the hotkey.
                elif hotkey is None:
                    if not comment_buf:
                        assert comment_block_start_line is None
                        comment_block_start_line = line_no
                    comment_buf.append(line)
                continue

            lines.append(line)
            while lines[-1][-1] == "\\":
                # Exclude the backslash character escaping the newline.
                lines[-1] = lines[-1][:-1]
                next_line = f.readline()
                # readline() returns the empty string upon EOF.
                if not next_line:
                    break
                lines.append(next_line.rstrip("\n"))
                line_no += 1

            if line.startswith((" ", "\t")):
                # command!
                if hotkey is None:
                    # Fatal error, so don't yield the exception.
                    raise MissingHotkeyError(
                        "Missing hotkey while reading a command: did you forget to escape a newline?",
                        line=line_no,
                    )
                command = lines.copy()
                command_start_line = line_block_start

                try:
                    keybind = Keybind(
                        hotkey,
                        command,
                        hotkey_start_line=hotkey_start_line,
                        command_start_line=command_start_line,
                        metadata=metadata,
                        hotkey_errors=hotkey_errors,
                    )
                except SXHKDParserError as e:
                    yield e
                else:
                    if section_handler is not None:
                        section_handler.current_section.add_keybind(keybind)
                    yield keybind

                hotkey = command = None
                hotkey_start_line = command_start_line = None
                lines.clear()
            else:
                # hotkey!
                assert hotkey is None and hotkey_start_line is None, repr(
                    (hotkey, hotkey_start_line, line)
                )
                hotkey = lines.copy()
                hotkey_start_line = line_block_start
                lines.clear()
                if comment_buf:
                    metadata = metadata_parser.parse(
                        comment_buf,
                        start_line=cast(int, comment_block_start_line),
                    )
                    comment_buf.clear()
                    comment_block_start_line = None
                else:
                    metadata = {}

    except Exception:
        raise
    finally:
        try:
            if section_handler is not None:
                section_handler.push_eof(line_no)
        finally:
            if close_io:
                f.close()



# flake8: noqa

"""Standardised keysyms."""

from typing import Set

__all__ = ["KEYSYMS"]

KEYSYMS: Set[str] = {
    "0",
    "1",
    "2",
    "3",
    "3270_AltCursor",
    "3270_Attn",
    "3270_BackTab",
    "3270_ChangeScreen",
    "3270_Copy",
    "3270_CursorBlink",
    "3270_CursorSelect",
    "3270_DeleteWord",
    "3270_Duplicate",
    "3270_Enter",
    "3270_EraseEOF",
    "3270_EraseInput",
    "3270_ExSelect",
    "3270_FieldMark",
    "3270_Ident",
    "3270_Jump",
    "3270_KeyClick",
    "3270_Left2",
    "3270_PA1",
    "3270_PA2",
    "3270_PA3",
    "3270_Play",
    "3270_PrintScreen",
    "3270_Quit",
    "3270_Record",
    "3270_Reset",
    "3270_Right2",
    "3270_Rule",
    "3270_Setup",
    "3270_Test",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    "a",
    "A",
    "aacute",
    "Aacute",
    "abelowdot",
    "Abelowdot",
    "abovedot",
    "abreve",
    "Abreve",
    "abreveacute",
    "Abreveacute",
    "abrevebelowdot",
    "Abrevebelowdot",
    "abrevegrave",
    "Abrevegrave",
    "abrevehook",
    "Abrevehook",
    "abrevetilde",
    "Abrevetilde",
    "AccessX_Enable",
    "AccessX_Feedback_Enable",
    "acircumflex",
    "Acircumflex",
    "acircumflexacute",
    "Acircumflexacute",
    "acircumflexbelowdot",
    "Acircumflexbelowdot",
    "acircumflexgrave",
    "Acircumflexgrave",
    "acircumflexhook",
    "Acircumflexhook",
    "acircumflextilde",
    "Acircumflextilde",
    "acute",
    "adiaeresis",
    "Adiaeresis",
    "ae",
    "AE",
    "agrave",
    "Agrave",
    "ahook",
    "Ahook",
    "Alt_L",
    "Alt_R",
    "amacron",
    "Amacron",
    "ampersand",
    "aogonek",
    "Aogonek",
    "apostrophe",
    "approxeq",
    "approximate",
    "Arabic_0",
    "Arabic_1",
    "Arabic_2",
    "Arabic_3",
    "Arabic_4",
    "Arabic_5",
    "Arabic_6",
    "Arabic_7",
    "Arabic_8",
    "Arabic_9",
    "Arabic_ain",
    "Arabic_alef",
    "Arabic_alefmaksura",
    "Arabic_beh",
    "Arabic_comma",
    "Arabic_dad",
    "Arabic_dal",
    "Arabic_damma",
    "Arabic_dammatan",
    "Arabic_ddal",
    "Arabic_farsi_yeh",
    "Arabic_fatha",
    "Arabic_fathatan",
    "Arabic_feh",
    "Arabic_fullstop",
    "Arabic_gaf",
    "Arabic_ghain",
    "Arabic_ha",
    "Arabic_hah",
    "Arabic_hamza",
    "Arabic_hamza_above",
    "Arabic_hamza_below",
    "Arabic_hamzaonalef",
    "Arabic_hamzaonwaw",
    "Arabic_hamzaonyeh",
    "Arabic_hamzaunderalef",
    "Arabic_heh",
    "Arabic_heh_doachashmee",
    "Arabic_heh_goal",
    "Arabic_jeem",
    "Arabic_jeh",
    "Arabic_kaf",
    "Arabic_kasra",
    "Arabic_kasratan",
    "Arabic_keheh",
    "Arabic_khah",
    "Arabic_lam",
    "Arabic_madda_above",
    "Arabic_maddaonalef",
    "Arabic_meem",
    "Arabic_noon",
    "Arabic_noon_ghunna",
    "Arabic_peh",
    "Arabic_percent",
    "Arabic_qaf",
    "Arabic_question_mark",
    "Arabic_ra",
    "Arabic_rreh",
    "Arabic_sad",
    "Arabic_seen",
    "Arabic_semicolon",
    "Arabic_shadda",
    "Arabic_sheen",
    "Arabic_sukun",
    "Arabic_superscript_alef",
    "Arabic_switch",
    "Arabic_tah",
    "Arabic_tatweel",
    "Arabic_tcheh",
    "Arabic_teh",
    "Arabic_tehmarbuta",
    "Arabic_thal",
    "Arabic_theh",
    "Arabic_tteh",
    "Arabic_veh",
    "Arabic_waw",
    "Arabic_yeh",
    "Arabic_yeh_baree",
    "Arabic_zah",
    "Arabic_zain",
    "aring",
    "Aring",
    "Armenian_accent",
    "Armenian_amanak",
    "Armenian_apostrophe",
    "Armenian_at",
    "Armenian_AT",
    "Armenian_ayb",
    "Armenian_AYB",
    "Armenian_ben",
    "Armenian_BEN",
    "Armenian_but",
    "Armenian_cha",
    "Armenian_CHA",
    "Armenian_da",
    "Armenian_DA",
    "Armenian_dza",
    "Armenian_DZA",
    "Armenian_e",
    "Armenian_E",
    "Armenian_exclam",
    "Armenian_fe",
    "Armenian_FE",
    "Armenian_full_stop",
    "Armenian_ghat",
    "Armenian_GHAT",
    "Armenian_gim",
    "Armenian_GIM",
    "Armenian_hi",
    "Armenian_HI",
    "Armenian_ho",
    "Armenian_HO",
    "Armenian_hyphen",
    "Armenian_ini",
    "Armenian_INI",
    "Armenian_je",
    "Armenian_JE",
    "Armenian_ke",
    "Armenian_KE",
    "Armenian_ken",
    "Armenian_KEN",
    "Armenian_khe",
    "Armenian_KHE",
    "Armenian_ligature_ew",
    "Armenian_lyun",
    "Armenian_LYUN",
    "Armenian_men",
    "Armenian_MEN",
    "Armenian_nu",
    "Armenian_NU",
    "Armenian_o",
    "Armenian_O",
    "Armenian_paruyk",
    "Armenian_pe",
    "Armenian_PE",
    "Armenian_pyur",
    "Armenian_PYUR",
    "Armenian_question",
    "Armenian_ra",
    "Armenian_RA",
    "Armenian_re",
    "Armenian_RE",
    "Armenian_se",
    "Armenian_SE",
    "Armenian_separation_mark",
    "Armenian_sha",
    "Armenian_SHA",
    "Armenian_shesht",
    "Armenian_tche",
    "Armenian_TCHE",
    "Armenian_to",
    "Armenian_TO",
    "Armenian_tsa",
    "Armenian_TSA",
    "Armenian_tso",
    "Armenian_TSO",
    "Armenian_tyun",
    "Armenian_TYUN",
    "Armenian_verjaket",
    "Armenian_vev",
    "Armenian_VEV",
    "Armenian_vo",
    "Armenian_VO",
    "Armenian_vyun",
    "Armenian_VYUN",
    "Armenian_yech",
    "Armenian_YECH",
    "Armenian_yentamna",
    "Armenian_za",
    "Armenian_ZA",
    "Armenian_zhe",
    "Armenian_ZHE",
    "asciicircum",
    "asciitilde",
    "asterisk",
    "at",
    "atilde",
    "Atilde",
    "AudibleBell_Enable",
    "b",
    "B",
    "babovedot",
    "Babovedot",
    "backslash",
    "BackSpace",
    "ballotcross",
    "bar",
    "because",
    "Begin",
    "blank",
    "botintegral",
    "botleftparens",
    "botleftsqbracket",
    "botleftsummation",
    "botrightparens",
    "botrightsqbracket",
    "botrightsummation",
    "bott",
    "botvertsummationconnector",
    "BounceKeys_Enable",
    "braceleft",
    "braceright",
    "bracketleft",
    "bracketright",
    "braille_blank",
    "braille_dot_1",
    "braille_dot_10",
    "braille_dot_2",
    "braille_dot_3",
    "braille_dot_4",
    "braille_dot_5",
    "braille_dot_6",
    "braille_dot_7",
    "braille_dot_8",
    "braille_dot_9",
    "braille_dots_1",
    "braille_dots_12",
    "braille_dots_123",
    "braille_dots_1234",
    "braille_dots_12345",
    "braille_dots_123456",
    "braille_dots_1234567",
    "braille_dots_12345678",
    "braille_dots_1234568",
    "braille_dots_123457",
    "braille_dots_1234578",
    "braille_dots_123458",
    "braille_dots_12346",
    "braille_dots_123467",
    "braille_dots_1234678",
    "braille_dots_123468",
    "braille_dots_12347",
    "braille_dots_123478",
    "braille_dots_12348",
    "braille_dots_1235",
    "braille_dots_12356",
    "braille_dots_123567",
    "braille_dots_1235678",
    "braille_dots_123568",
    "braille_dots_12357",
    "braille_dots_123578",
    "braille_dots_12358",
    "braille_dots_1236",
    "braille_dots_12367",
    "braille_dots_123678",
    "braille_dots_12368",
    "braille_dots_1237",
    "braille_dots_12378",
    "braille_dots_1238",
    "braille_dots_124",
    "braille_dots_1245",
    "braille_dots_12456",
    "braille_dots_124567",
    "braille_dots_1245678",
    "braille_dots_124568",
    "braille_dots_12457",
    "braille_dots_124578",
    "braille_dots_12458",
    "braille_dots_1246",
    "braille_dots_12467",
    "braille_dots_124678",
    "braille_dots_12468",
    "braille_dots_1247",
    "braille_dots_12478",
    "braille_dots_1248",
    "braille_dots_125",
    "braille_dots_1256",
    "braille_dots_12567",
    "braille_dots_125678",
    "braille_dots_12568",
    "braille_dots_1257",
    "braille_dots_12578",
    "braille_dots_1258",
    "braille_dots_126",
    "braille_dots_1267",
    "braille_dots_12678",
    "braille_dots_1268",
    "braille_dots_127",
    "braille_dots_1278",
    "braille_dots_128",
    "braille_dots_13",
    "braille_dots_134",
    "braille_dots_1345",
    "braille_dots_13456",
    "braille_dots_134567",
    "braille_dots_1345678",
    "braille_dots_134568",
    "braille_dots_13457",
    "braille_dots_134578",
    "braille_dots_13458",
    "braille_dots_1346",
    "braille_dots_13467",
    "braille_dots_134678",
    "braille_dots_13468",
    "braille_dots_1347",
    "braille_dots_13478",
    "braille_dots_1348",
    "braille_dots_135",
    "braille_dots_1356",
    "braille_dots_13567",
    "braille_dots_135678",
    "braille_dots_13568",
    "braille_dots_1357",
    "braille_dots_13578",
    "braille_dots_1358",
    "braille_dots_136",
    "braille_dots_1367",
    "braille_dots_13678",
    "braille_dots_1368",
    "braille_dots_137",
    "braille_dots_1378",
    "braille_dots_138",
    "braille_dots_14",
    "braille_dots_145",
    "braille_dots_1456",
    "braille_dots_14567",
    "braille_dots_145678",
    "braille_dots_14568",
    "braille_dots_1457",
    "braille_dots_14578",
    "braille_dots_1458",
    "braille_dots_146",
    "braille_dots_1467",
    "braille_dots_14678",
    "braille_dots_1468",
    "braille_dots_147",
    "braille_dots_1478",
    "braille_dots_148",
    "braille_dots_15",
    "braille_dots_156",
    "braille_dots_1567",
    "braille_dots_15678",
    "braille_dots_1568",
    "braille_dots_157",
    "braille_dots_1578",
    "braille_dots_158",
    "braille_dots_16",
    "braille_dots_167",
    "braille_dots_1678",
    "braille_dots_168",
    "braille_dots_17",
    "braille_dots_178",
    "braille_dots_18",
    "braille_dots_2",
    "braille_dots_23",
    "braille_dots_234",
    "braille_dots_2345",
    "braille_dots_23456",
    "braille_dots_234567",
    "braille_dots_2345678",
    "braille_dots_234568",
    "braille_dots_23457",
    "braille_dots_234578",
    "braille_dots_23458",
    "braille_dots_2346",
    "braille_dots_23467",
    "braille_dots_234678",
    "braille_dots_23468",
    "braille_dots_2347",
    "braille_dots_23478",
    "braille_dots_2348",
    "braille_dots_235",
    "braille_dots_2356",
    "braille_dots_23567",
    "braille_dots_235678",
    "braille_dots_23568",
    "braille_dots_2357",
    "braille_dots_23578",
    "braille_dots_2358",
    "braille_dots_236",
    "braille_dots_2367",
    "braille_dots_23678",
    "braille_dots_2368",
    "braille_dots_237",
    "braille_dots_2378",
    "braille_dots_238",
    "braille_dots_24",
    "braille_dots_245",
    "braille_dots_2456",
    "braille_dots_24567",
    "braille_dots_245678",
    "braille_dots_24568",
    "braille_dots_2457",
    "braille_dots_24578",
    "braille_dots_2458",
    "braille_dots_246",
    "braille_dots_2467",
    "braille_dots_24678",
    "braille_dots_2468",
    "braille_dots_247",
    "braille_dots_2478",
    "braille_dots_248",
    "braille_dots_25",
    "braille_dots_256",
    "braille_dots_2567",
    "braille_dots_25678",
    "braille_dots_2568",
    "braille_dots_257",
    "braille_dots_2578",
    "braille_dots_258",
    "braille_dots_26",
    "braille_dots_267",
    "braille_dots_2678",
    "braille_dots_268",
    "braille_dots_27",
    "braille_dots_278",
    "braille_dots_28",
    "braille_dots_3",
    "braille_dots_34",
    "braille_dots_345",
    "braille_dots_3456",
    "braille_dots_34567",
    "braille_dots_345678",
    "braille_dots_34568",
    "braille_dots_3457",
    "braille_dots_34578",
    "braille_dots_3458",
    "braille_dots_346",
    "braille_dots_3467",
    "braille_dots_34678",
    "braille_dots_3468",
    "braille_dots_347",
    "braille_dots_3478",
    "braille_dots_348",
    "braille_dots_35",
    "braille_dots_356",
    "braille_dots_3567",
    "braille_dots_35678",
    "braille_dots_3568",
    "braille_dots_357",
    "braille_dots_3578",
    "braille_dots_358",
    "braille_dots_36",
    "braille_dots_367",
    "braille_dots_3678",
    "braille_dots_368",
    "braille_dots_37",
    "braille_dots_378",
    "braille_dots_38",
    "braille_dots_4",
    "braille_dots_45",
    "braille_dots_456",
    "braille_dots_4567",
    "braille_dots_45678",
    "braille_dots_4568",
    "braille_dots_457",
    "braille_dots_4578",
    "braille_dots_458",
    "braille_dots_46",
    "braille_dots_467",
    "braille_dots_4678",
    "braille_dots_468",
    "braille_dots_47",
    "braille_dots_478",
    "braille_dots_48",
    "braille_dots_5",
    "braille_dots_56",
    "braille_dots_567",
    "braille_dots_5678",
    "braille_dots_568",
    "braille_dots_57",
    "braille_dots_578",
    "braille_dots_58",
    "braille_dots_6",
    "braille_dots_67",
    "braille_dots_678",
    "braille_dots_68",
    "braille_dots_7",
    "braille_dots_78",
    "braille_dots_8",
    "Break",
    "breve",
    "brokenbar",
    "Byelorussian_shortu",
    "Byelorussian_SHORTU",
    "c",
    "C",
    "cabovedot",
    "Cabovedot",
    "cacute",
    "Cacute",
    "Cancel",
    "Caps_Lock",
    "careof",
    "caret",
    "caron",
    "ccaron",
    "Ccaron",
    "ccedilla",
    "Ccedilla",
    "ccircumflex",
    "Ccircumflex",
    "cedilla",
    "cent",
    "c_h",
    "ch",
    "C_h",
    "Ch",
    "C_H",
    "CH",
    "checkerboard",
    "checkmark",
    "circle",
    "Clear",
    "club",
    "Codeinput",
    "colon",
    "ColonSign",
    "comma",
    "containsas",
    "Control_L",
    "Control_R",
    "copyright",
    "cr",
    "crossinglines",
    "CruzeiroSign",
    "cuberoot",
    "currency",
    "cursor",
    "Cyrillic_a",
    "Cyrillic_A",
    "Cyrillic_be",
    "Cyrillic_BE",
    "Cyrillic_che",
    "Cyrillic_CHE",
    "Cyrillic_che_descender",
    "Cyrillic_CHE_descender",
    "Cyrillic_che_vertstroke",
    "Cyrillic_CHE_vertstroke",
    "Cyrillic_de",
    "Cyrillic_DE",
    "Cyrillic_dzhe",
    "Cyrillic_DZHE",
    "Cyrillic_e",
    "Cyrillic_E",
    "Cyrillic_ef",
    "Cyrillic_EF",
    "Cyrillic_el",
    "Cyrillic_EL",
    "Cyrillic_em",
    "Cyrillic_EM",
    "Cyrillic_en",
    "Cyrillic_EN",
    "Cyrillic_en_descender",
    "Cyrillic_EN_descender",
    "Cyrillic_er",
    "Cyrillic_ER",
    "Cyrillic_es",
    "Cyrillic_ES",
    "Cyrillic_ghe",
    "Cyrillic_GHE",
    "Cyrillic_ghe_bar",
    "Cyrillic_GHE_bar",
    "Cyrillic_ha",
    "Cyrillic_HA",
    "Cyrillic_ha_descender",
    "Cyrillic_HA_descender",
    "Cyrillic_hardsign",
    "Cyrillic_HARDSIGN",
    "Cyrillic_i",
    "Cyrillic_I",
    "Cyrillic_ie",
    "Cyrillic_IE",
    "Cyrillic_i_macron",
    "Cyrillic_I_macron",
    "Cyrillic_io",
    "Cyrillic_IO",
    "Cyrillic_je",
    "Cyrillic_JE",
    "Cyrillic_ka",
    "Cyrillic_KA",
    "Cyrillic_ka_descender",
    "Cyrillic_KA_descender",
    "Cyrillic_ka_vertstroke",
    "Cyrillic_KA_vertstroke",
    "Cyrillic_lje",
    "Cyrillic_LJE",
    "Cyrillic_nje",
    "Cyrillic_NJE",
    "Cyrillic_o",
    "Cyrillic_O",
    "Cyrillic_o_bar",
    "Cyrillic_O_bar",
    "Cyrillic_pe",
    "Cyrillic_PE",
    "Cyrillic_schwa",
    "Cyrillic_SCHWA",
    "Cyrillic_sha",
    "Cyrillic_SHA",
    "Cyrillic_shcha",
    "Cyrillic_SHCHA",
    "Cyrillic_shha",
    "Cyrillic_SHHA",
    "Cyrillic_shorti",
    "Cyrillic_SHORTI",
    "Cyrillic_softsign",
    "Cyrillic_SOFTSIGN",
    "Cyrillic_te",
    "Cyrillic_TE",
    "Cyrillic_tse",
    "Cyrillic_TSE",
    "Cyrillic_u",
    "Cyrillic_U",
    "Cyrillic_u_macron",
    "Cyrillic_U_macron",
    "Cyrillic_u_straight",
    "Cyrillic_U_straight",
    "Cyrillic_u_straight_bar",
    "Cyrillic_U_straight_bar",
    "Cyrillic_ve",
    "Cyrillic_VE",
    "Cyrillic_ya",
    "Cyrillic_YA",
    "Cyrillic_yeru",
    "Cyrillic_YERU",
    "Cyrillic_yu",
    "Cyrillic_YU",
    "Cyrillic_ze",
    "Cyrillic_ZE",
    "Cyrillic_zhe",
    "Cyrillic_ZHE",
    "Cyrillic_zhe_descender",
    "Cyrillic_ZHE_descender",
    "d",
    "D",
    "dabovedot",
    "Dabovedot",
    "dagger",
    "dcaron",
    "Dcaron",
    "dead_a",
    "dead_A",
    "dead_abovecomma",
    "dead_abovedot",
    "dead_abovereversedcomma",
    "dead_abovering",
    "dead_aboveverticalline",
    "dead_acute",
    "dead_belowbreve",
    "dead_belowcircumflex",
    "dead_belowcomma",
    "dead_belowdiaeresis",
    "dead_belowdot",
    "dead_belowmacron",
    "dead_belowring",
    "dead_belowtilde",
    "dead_belowverticalline",
    "dead_breve",
    "dead_capital_schwa",
    "dead_caron",
    "dead_cedilla",
    "dead_circumflex",
    "dead_currency",
    "dead_dasia",
    "dead_diaeresis",
    "dead_doubleacute",
    "dead_doublegrave",
    "dead_e",
    "dead_E",
    "dead_grave",
    "dead_greek",
    "dead_hook",
    "dead_horn",
    "dead_i",
    "dead_I",
    "dead_invertedbreve",
    "dead_iota",
    "dead_longsolidusoverlay",
    "dead_lowline",
    "dead_macron",
    "dead_o",
    "dead_O",
    "dead_ogonek",
    "dead_perispomeni",
    "dead_psili",
    "dead_semivoiced_sound",
    "dead_small_schwa",
    "dead_stroke",
    "dead_tilde",
    "dead_u",
    "dead_U",
    "dead_voiced_sound",
    "decimalpoint",
    "degree",
    "Delete",
    "diaeresis",
    "diamond",
    "digitspace",
    "dintegral",
    "division",
    "dollar",
    "DongSign",
    "doubbaselinedot",
    "doubleacute",
    "doubledagger",
    "doublelowquotemark",
    "Down",
    "downarrow",
    "downcaret",
    "downshoe",
    "downstile",
    "downtack",
    "dstroke",
    "Dstroke",
    "e",
    "E",
    "eabovedot",
    "Eabovedot",
    "eacute",
    "Eacute",
    "ebelowdot",
    "Ebelowdot",
    "ecaron",
    "Ecaron",
    "ecircumflex",
    "Ecircumflex",
    "ecircumflexacute",
    "Ecircumflexacute",
    "ecircumflexbelowdot",
    "Ecircumflexbelowdot",
    "ecircumflexgrave",
    "Ecircumflexgrave",
    "ecircumflexhook",
    "Ecircumflexhook",
    "ecircumflextilde",
    "Ecircumflextilde",
    "EcuSign",
    "ediaeresis",
    "Ediaeresis",
    "egrave",
    "Egrave",
    "ehook",
    "Ehook",
    "eightsubscript",
    "eightsuperior",
    "Eisu_Shift",
    "Eisu_toggle",
    "elementof",
    "ellipsis",
    "em3space",
    "em4space",
    "emacron",
    "Emacron",
    "emdash",
    "emfilledcircle",
    "emfilledrect",
    "emopencircle",
    "emopenrectangle",
    "emptyset",
    "emspace",
    "End",
    "endash",
    "enfilledcircbullet",
    "enfilledsqbullet",
    "eng",
    "ENG",
    "enopencircbullet",
    "enopensquarebullet",
    "enspace",
    "eogonek",
    "Eogonek",
    "equal",
    "Escape",
    "eth",
    "Eth",
    "ETH",
    "etilde",
    "Etilde",
    "EuroSign",
    "exclam",
    "exclamdown",
    "Execute",
    "ezh",
    "EZH",
    "f",
    "F",
    "F1",
    "F10",
    "F11",
    "F12",
    "F13",
    "F14",
    "F15",
    "F16",
    "F17",
    "F18",
    "F19",
    "F2",
    "F20",
    "F21",
    "F22",
    "F23",
    "F24",
    "F25",
    "F26",
    "F27",
    "F28",
    "F29",
    "F3",
    "F30",
    "F31",
    "F32",
    "F33",
    "F34",
    "F35",
    "F4",
    "F5",
    "F6",
    "F7",
    "F8",
    "F9",
    "fabovedot",
    "Fabovedot",
    "Farsi_0",
    "Farsi_1",
    "Farsi_2",
    "Farsi_3",
    "Farsi_4",
    "Farsi_5",
    "Farsi_6",
    "Farsi_7",
    "Farsi_8",
    "Farsi_9",
    "Farsi_yeh",
    "femalesymbol",
    "ff",
    "FFrancSign",
    "figdash",
    "filledlefttribullet",
    "filledrectbullet",
    "filledrighttribullet",
    "filledtribulletdown",
    "filledtribulletup",
    "Find",
    "First_Virtual_Screen",
    "fiveeighths",
    "fivesixths",
    "fivesubscript",
    "fivesuperior",
    "fourfifths",
    "foursubscript",
    "foursuperior",
    "fourthroot",
    "function",
    "g",
    "G",
    "gabovedot",
    "Gabovedot",
    "gbreve",
    "Gbreve",
    "gcaron",
    "Gcaron",
    "gcedilla",
    "Gcedilla",
    "gcircumflex",
    "Gcircumflex",
    "Georgian_an",
    "Georgian_ban",
    "Georgian_can",
    "Georgian_char",
    "Georgian_chin",
    "Georgian_cil",
    "Georgian_don",
    "Georgian_en",
    "Georgian_fi",
    "Georgian_gan",
    "Georgian_ghan",
    "Georgian_hae",
    "Georgian_har",
    "Georgian_he",
    "Georgian_hie",
    "Georgian_hoe",
    "Georgian_in",
    "Georgian_jhan",
    "Georgian_jil",
    "Georgian_kan",
    "Georgian_khar",
    "Georgian_las",
    "Georgian_man",
    "Georgian_nar",
    "Georgian_on",
    "Georgian_par",
    "Georgian_phar",
    "Georgian_qar",
    "Georgian_rae",
    "Georgian_san",
    "Georgian_shin",
    "Georgian_tan",
    "Georgian_tar",
    "Georgian_un",
    "Georgian_vin",
    "Georgian_we",
    "Georgian_xan",
    "Georgian_zen",
    "Georgian_zhar",
    "grave",
    "greater",
    "greaterthanequal",
    "Greek_accentdieresis",
    "Greek_alpha",
    "Greek_ALPHA",
    "Greek_alphaaccent",
    "Greek_ALPHAaccent",
    "Greek_beta",
    "Greek_BETA",
    "Greek_chi",
    "Greek_CHI",
    "Greek_delta",
    "Greek_DELTA",
    "Greek_epsilon",
    "Greek_EPSILON",
    "Greek_epsilonaccent",
    "Greek_EPSILONaccent",
    "Greek_eta",
    "Greek_ETA",
    "Greek_etaaccent",
    "Greek_ETAaccent",
    "Greek_finalsmallsigma",
    "Greek_gamma",
    "Greek_GAMMA",
    "Greek_horizbar",
    "Greek_iota",
    "Greek_IOTA",
    "Greek_iotaaccent",
    "Greek_IOTAaccent",
    "Greek_iotaaccentdieresis",
    "Greek_IOTAdiaeresis",
    "Greek_iotadieresis",
    "Greek_IOTAdieresis",
    "Greek_kappa",
    "Greek_KAPPA",
    "Greek_lambda",
    "Greek_LAMBDA",
    "Greek_lamda",
    "Greek_LAMDA",
    "Greek_mu",
    "Greek_MU",
    "Greek_nu",
    "Greek_NU",
    "Greek_omega",
    "Greek_OMEGA",
    "Greek_omegaaccent",
    "Greek_OMEGAaccent",
    "Greek_omicron",
    "Greek_OMICRON",
    "Greek_omicronaccent",
    "Greek_OMICRONaccent",
    "Greek_phi",
    "Greek_PHI",
    "Greek_pi",
    "Greek_PI",
    "Greek_psi",
    "Greek_PSI",
    "Greek_rho",
    "Greek_RHO",
    "Greek_sigma",
    "Greek_SIGMA",
    "Greek_switch",
    "Greek_tau",
    "Greek_TAU",
    "Greek_theta",
    "Greek_THETA",
    "Greek_upsilon",
    "Greek_UPSILON",
    "Greek_upsilonaccent",
    "Greek_UPSILONaccent",
    "Greek_upsilonaccentdieresis",
    "Greek_upsilondieresis",
    "Greek_UPSILONdieresis",
    "Greek_xi",
    "Greek_XI",
    "Greek_zeta",
    "Greek_ZETA",
    "guillemotleft",
    "guillemotright",
    "h",
    "H",
    "hairspace",
    "Hangul",
    "Hangul_A",
    "Hangul_AE",
    "Hangul_AraeA",
    "Hangul_AraeAE",
    "Hangul_Banja",
    "Hangul_Cieuc",
    "Hangul_Codeinput",
    "Hangul_Dikeud",
    "Hangul_E",
    "Hangul_End",
    "Hangul_EO",
    "Hangul_EU",
    "Hangul_Hanja",
    "Hangul_Hieuh",
    "Hangul_I",
    "Hangul_Ieung",
    "Hangul_Jamo",
    "Hangul_J_Cieuc",
    "Hangul_J_Dikeud",
    "Hangul_Jeonja",
    "Hangul_J_Hieuh",
    "Hangul_Jieuj",
    "Hangul_J_Ieung",
    "Hangul_J_Jieuj",
    "Hangul_J_Khieuq",
    "Hangul_J_Kiyeog",
    "Hangul_J_KiyeogSios",
    "Hangul_J_KkogjiDalrinIeung",
    "Hangul_J_Mieum",
    "Hangul_J_Nieun",
    "Hangul_J_NieunHieuh",
    "Hangul_J_NieunJieuj",
    "Hangul_J_PanSios",
    "Hangul_J_Phieuf",
    "Hangul_J_Pieub",
    "Hangul_J_PieubSios",
    "Hangul_J_Rieul",
    "Hangul_J_RieulHieuh",
    "Hangul_J_RieulKiyeog",
    "Hangul_J_RieulMieum",
    "Hangul_J_RieulPhieuf",
    "Hangul_J_RieulPieub",
    "Hangul_J_RieulSios",
    "Hangul_J_RieulTieut",
    "Hangul_J_Sios",
    "Hangul_J_SsangKiyeog",
    "Hangul_J_SsangSios",
    "Hangul_J_Tieut",
    "Hangul_J_YeorinHieuh",
    "Hangul_Khieuq",
    "Hangul_Kiyeog",
    "Hangul_KiyeogSios",
    "Hangul_KkogjiDalrinIeung",
    "Hangul_Mieum",
    "Hangul_MultipleCandidate",
    "Hangul_Nieun",
    "Hangul_NieunHieuh",
    "Hangul_NieunJieuj",
    "Hangul_O",
    "Hangul_OE",
    "Hangul_PanSios",
    "Hangul_Phieuf",
    "Hangul_Pieub",
    "Hangul_PieubSios",
    "Hangul_PostHanja",
    "Hangul_PreHanja",
    "Hangul_PreviousCandidate",
    "Hangul_Rieul",
    "Hangul_RieulHieuh",
    "Hangul_RieulKiyeog",
    "Hangul_RieulMieum",
    "Hangul_RieulPhieuf",
    "Hangul_RieulPieub",
    "Hangul_RieulSios",
    "Hangul_RieulTieut",
    "Hangul_RieulYeorinHieuh",
    "Hangul_Romaja",
    "Hangul_SingleCandidate",
    "Hangul_Sios",
    "Hangul_Special",
    "Hangul_SsangDikeud",
    "Hangul_SsangJieuj",
    "Hangul_SsangKiyeog",
    "Hangul_SsangPieub",
    "Hangul_SsangSios",
    "Hangul_Start",
    "Hangul_SunkyeongeumMieum",
    "Hangul_SunkyeongeumPhieuf",
    "Hangul_SunkyeongeumPieub",
    "Hangul_switch",
    "Hangul_Tieut",
    "Hangul_U",
    "Hangul_WA",
    "Hangul_WAE",
    "Hangul_WE",
    "Hangul_WEO",
    "Hangul_WI",
    "Hangul_YA",
    "Hangul_YAE",
    "Hangul_YE",
    "Hangul_YEO",
    "Hangul_YeorinHieuh",
    "Hangul_YI",
    "Hangul_YO",
    "Hangul_YU",
    "Hankaku",
    "hcircumflex",
    "Hcircumflex",
    "heart",
    "hebrew_aleph",
    "hebrew_ayin",
    "hebrew_bet",
    "hebrew_beth",
    "hebrew_chet",
    "hebrew_dalet",
    "hebrew_daleth",
    "hebrew_doublelowline",
    "hebrew_finalkaph",
    "hebrew_finalmem",
    "hebrew_finalnun",
    "hebrew_finalpe",
    "hebrew_finalzade",
    "hebrew_finalzadi",
    "hebrew_gimel",
    "hebrew_gimmel",
    "hebrew_he",
    "hebrew_het",
    "hebrew_kaph",
    "hebrew_kuf",
    "hebrew_lamed",
    "hebrew_mem",
    "hebrew_nun",
    "hebrew_pe",
    "hebrew_qoph",
    "hebrew_resh",
    "hebrew_samech",
    "hebrew_samekh",
    "hebrew_shin",
    "Hebrew_switch",
    "hebrew_taf",
    "hebrew_taw",
    "hebrew_tet",
    "hebrew_teth",
    "hebrew_waw",
    "hebrew_yod",
    "hebrew_zade",
    "hebrew_zadi",
    "hebrew_zain",
    "hebrew_zayin",
    "Help",
    "Henkan",
    "Henkan_Mode",
    "hexagram",
    "Hiragana",
    "Hiragana_Katakana",
    "Home",
    "horizconnector",
    "horizlinescan1",
    "horizlinescan3",
    "horizlinescan5",
    "horizlinescan7",
    "horizlinescan9",
    "hstroke",
    "Hstroke",
    "ht",
    "Hyper_L",
    "Hyper_R",
    "hyphen",
    "i",
    "I",
    "Iabovedot",
    "iacute",
    "Iacute",
    "ibelowdot",
    "Ibelowdot",
    "ibreve",
    "Ibreve",
    "icircumflex",
    "Icircumflex",
    "identical",
    "idiaeresis",
    "Idiaeresis",
    "idotless",
    "ifonlyif",
    "igrave",
    "Igrave",
    "ihook",
    "Ihook",
    "imacron",
    "Imacron",
    "implies",
    "includedin",
    "includes",
    "infinity",
    "Insert",
    "integral",
    "intersection",
    "iogonek",
    "Iogonek",
    "ISO_Center_Object",
    "ISO_Continuous_Underline",
    "ISO_Discontinuous_Underline",
    "ISO_Emphasize",
    "ISO_Enter",
    "ISO_Fast_Cursor_Down",
    "ISO_Fast_Cursor_Left",
    "ISO_Fast_Cursor_Right",
    "ISO_Fast_Cursor_Up",
    "ISO_First_Group",
    "ISO_First_Group_Lock",
    "ISO_Group_Latch",
    "ISO_Group_Lock",
    "ISO_Group_Shift",
    "ISO_Last_Group",
    "ISO_Last_Group_Lock",
    "ISO_Left_Tab",
    "ISO_Level2_Latch",
    "ISO_Level3_Latch",
    "ISO_Level3_Lock",
    "ISO_Level3_Shift",
    "ISO_Level5_Latch",
    "ISO_Level5_Lock",
    "ISO_Level5_Shift",
    "ISO_Lock",
    "ISO_Move_Line_Down",
    "ISO_Move_Line_Up",
    "ISO_Next_Group",
    "ISO_Next_Group_Lock",
    "ISO_Partial_Line_Down",
    "ISO_Partial_Line_Up",
    "ISO_Partial_Space_Left",
    "ISO_Partial_Space_Right",
    "ISO_Prev_Group",
    "ISO_Prev_Group_Lock",
    "ISO_Release_Both_Margins",
    "ISO_Release_Margin_Left",
    "ISO_Release_Margin_Right",
    "ISO_Set_Margin_Left",
    "ISO_Set_Margin_Right",
    "itilde",
    "Itilde",
    "j",
    "J",
    "jcircumflex",
    "Jcircumflex",
    "jot",
    "k",
    "K",
    "kana_a",
    "kana_A",
    "kana_CHI",
    "kana_closingbracket",
    "kana_comma",
    "kana_conjunctive",
    "kana_e",
    "kana_E",
    "kana_FU",
    "kana_fullstop",
    "kana_HA",
    "kana_HE",
    "kana_HI",
    "kana_HO",
    "kana_HU",
    "kana_i",
    "kana_I",
    "kana_KA",
    "kana_KE",
    "kana_KI",
    "kana_KO",
    "kana_KU",
    "Kana_Lock",
    "kana_MA",
    "kana_ME",
    "kana_MI",
    "kana_middledot",
    "kana_MO",
    "kana_MU",
    "kana_N",
    "kana_NA",
    "kana_NE",
    "kana_NI",
    "kana_NO",
    "kana_NU",
    "kana_o",
    "kana_O",
    "kana_openingbracket",
    "kana_RA",
    "kana_RE",
    "kana_RI",
    "kana_RO",
    "kana_RU",
    "kana_SA",
    "kana_SE",
    "kana_SHI",
    "Kana_Shift",
    "kana_SO",
    "kana_SU",
    "kana_switch",
    "kana_TA",
    "kana_TE",
    "kana_TI",
    "kana_TO",
    "kana_tsu",
    "kana_TSU",
    "kana_tu",
    "kana_TU",
    "kana_u",
    "kana_U",
    "kana_WA",
    "kana_WO",
    "kana_ya",
    "kana_YA",
    "kana_yo",
    "kana_YO",
    "kana_yu",
    "kana_YU",
    "Kanji",
    "Kanji_Bangou",
    "kappa",
    "Katakana",
    "kcedilla",
    "Kcedilla",
    "Korean_Won",
    "KP_0",
    "KP_1",
    "KP_2",
    "KP_3",
    "KP_4",
    "KP_5",
    "KP_6",
    "KP_7",
    "KP_8",
    "KP_9",
    "KP_Add",
    "KP_Begin",
    "KP_Decimal",
    "KP_Delete",
    "KP_Divide",
    "KP_Down",
    "KP_End",
    "KP_Enter",
    "KP_Equal",
    "KP_F1",
    "KP_F2",
    "KP_F3",
    "KP_F4",
    "KP_Home",
    "KP_Insert",
    "KP_Left",
    "KP_Multiply",
    "KP_Next",
    "KP_Page_Down",
    "KP_Page_Up",
    "KP_Prior",
    "KP_Right",
    "KP_Separator",
    "KP_Space",
    "KP_Subtract",
    "KP_Tab",
    "KP_Up",
    "kra",
    "l",
    "L",
    "L1",
    "L10",
    "L2",
    "L3",
    "L4",
    "L5",
    "L6",
    "L7",
    "L8",
    "L9",
    "lacute",
    "Lacute",
    "Last_Virtual_Screen",
    "latincross",
    "lbelowdot",
    "Lbelowdot",
    "lcaron",
    "Lcaron",
    "lcedilla",
    "Lcedilla",
    "Left",
    "leftanglebracket",
    "leftarrow",
    "leftcaret",
    "leftdoublequotemark",
    "leftmiddlecurlybrace",
    "leftopentriangle",
    "leftpointer",
    "leftradical",
    "leftshoe",
    "leftsinglequotemark",
    "leftt",
    "lefttack",
    "less",
    "lessthanequal",
    "lf",
    "Linefeed",
    "LiraSign",
    "logicaland",
    "logicalor",
    "lowleftcorner",
    "lowrightcorner",
    "lstroke",
    "Lstroke",
    "m",
    "M",
    "mabovedot",
    "Mabovedot",
    "Macedonia_dse",
    "Macedonia_DSE",
    "Macedonia_gje",
    "Macedonia_GJE",
    "Macedonia_kje",
    "Macedonia_KJE",
    "macron",
    "Mae_Koho",
    "malesymbol",
    "maltesecross",
    "marker",
    "masculine",
    "Massyo",
    "Menu",
    "Meta_L",
    "Meta_R",
    "MillSign",
    "minus",
    "minutes",
    "Mode_switch",
    "MouseKeys_Accel_Enable",
    "MouseKeys_Enable",
    "mu",
    "Muhenkan",
    "Multi_key",
    "MultipleCandidate",
    "multiply",
    "musicalflat",
    "musicalsharp",
    "n",
    "N",
    "nabla",
    "nacute",
    "Nacute",
    "NairaSign",
    "ncaron",
    "Ncaron",
    "ncedilla",
    "Ncedilla",
    "NewSheqelSign",
    "Next",
    "Next_Virtual_Screen",
    "ninesubscript",
    "ninesuperior",
    "nl",
    "nobreakspace",
    "notapproxeq",
    "notelementof",
    "notequal",
    "notidentical",
    "notsign",
    "ntilde",
    "Ntilde",
    "numbersign",
    "numerosign",
    "Num_Lock",
    "o",
    "O",
    "oacute",
    "Oacute",
    "obarred",
    "Obarred",
    "obelowdot",
    "Obelowdot",
    "ocaron",
    "Ocaron",
    "ocircumflex",
    "Ocircumflex",
    "ocircumflexacute",
    "Ocircumflexacute",
    "ocircumflexbelowdot",
    "Ocircumflexbelowdot",
    "ocircumflexgrave",
    "Ocircumflexgrave",
    "ocircumflexhook",
    "Ocircumflexhook",
    "ocircumflextilde",
    "Ocircumflextilde",
    "odiaeresis",
    "Odiaeresis",
    "odoubleacute",
    "Odoubleacute",
    "oe",
    "OE",
    "ogonek",
    "ograve",
    "Ograve",
    "ohook",
    "Ohook",
    "ohorn",
    "Ohorn",
    "ohornacute",
    "Ohornacute",
    "ohornbelowdot",
    "Ohornbelowdot",
    "ohorngrave",
    "Ohorngrave",
    "ohornhook",
    "Ohornhook",
    "ohorntilde",
    "Ohorntilde",
    "omacron",
    "Omacron",
    "oneeighth",
    "onefifth",
    "onehalf",
    "onequarter",
    "onesixth",
    "onesubscript",
    "onesuperior",
    "onethird",
    "ooblique",
    "Ooblique",
    "openrectbullet",
    "openstar",
    "opentribulletdown",
    "opentribulletup",
    "ordfeminine",
    "oslash",
    "Oslash",
    "otilde",
    "Otilde",
    "overbar",
    "Overlay1_Enable",
    "Overlay2_Enable",
    "overline",
    "p",
    "P",
    "pabovedot",
    "Pabovedot",
    "Page_Down",
    "Page_Up",
    "paragraph",
    "parenleft",
    "parenright",
    "partdifferential",
    "partialderivative",
    "Pause",
    "percent",
    "period",
    "periodcentered",
    "permille",
    "PesetaSign",
    "phonographcopyright",
    "plus",
    "plusminus",
    "Pointer_Accelerate",
    "Pointer_Button1",
    "Pointer_Button2",
    "Pointer_Button3",
    "Pointer_Button4",
    "Pointer_Button5",
    "Pointer_Button_Dflt",
    "Pointer_DblClick1",
    "Pointer_DblClick2",
    "Pointer_DblClick3",
    "Pointer_DblClick4",
    "Pointer_DblClick5",
    "Pointer_DblClick_Dflt",
    "Pointer_DfltBtnNext",
    "Pointer_DfltBtnPrev",
    "Pointer_Down",
    "Pointer_DownLeft",
    "Pointer_DownRight",
    "Pointer_Drag1",
    "Pointer_Drag2",
    "Pointer_Drag3",
    "Pointer_Drag4",
    "Pointer_Drag5",
    "Pointer_Drag_Dflt",
    "Pointer_EnableKeys",
    "Pointer_Left",
    "Pointer_Right",
    "Pointer_Up",
    "Pointer_UpLeft",
    "Pointer_UpRight",
    "prescription",
    "PreviousCandidate",
    "Prev_Virtual_Screen",
    "Print",
    "Prior",
    "prolongedsound",
    "punctspace",
    "q",
    "Q",
    "quad",
    "question",
    "questiondown",
    "quotedbl",
    "quoteleft",
    "quoteright",
    "r",
    "R",
    "R1",
    "R10",
    "R11",
    "R12",
    "R13",
    "R14",
    "R15",
    "R2",
    "R3",
    "R4",
    "R5",
    "R6",
    "R7",
    "R8",
    "R9",
    "racute",
    "Racute",
    "radical",
    "rcaron",
    "Rcaron",
    "rcedilla",
    "Rcedilla",
    "Redo",
    "registered",
    "RepeatKeys_Enable",
    "Return",
    "Right",
    "rightanglebracket",
    "rightarrow",
    "rightcaret",
    "rightdoublequotemark",
    "rightmiddlecurlybrace",
    "rightmiddlesummation",
    "rightopentriangle",
    "rightpointer",
    "rightshoe",
    "rightsinglequotemark",
    "rightt",
    "righttack",
    "Romaji",
    "RupeeSign",
    "s",
    "S",
    "sabovedot",
    "Sabovedot",
    "sacute",
    "Sacute",
    "scaron",
    "Scaron",
    "scedilla",
    "Scedilla",
    "schwa",
    "SCHWA",
    "scircumflex",
    "Scircumflex",
    "script_switch",
    "Scroll_Lock",
    "seconds",
    "section",
    "Select",
    "semicolon",
    "semivoicedsound",
    "Serbian_dje",
    "Serbian_DJE",
    "Serbian_dze",
    "Serbian_DZE",
    "Serbian_je",
    "Serbian_JE",
    "Serbian_lje",
    "Serbian_LJE",
    "Serbian_nje",
    "Serbian_NJE",
    "Serbian_tshe",
    "Serbian_TSHE",
    "seveneighths",
    "sevensubscript",
    "sevensuperior",
    "Shift_L",
    "Shift_Lock",
    "Shift_R",
    "signaturemark",
    "signifblank",
    "similarequal",
    "SingleCandidate",
    "singlelowquotemark",
    "Sinh_a",
    "Sinh_aa",
    "Sinh_aa2",
    "Sinh_ae",
    "Sinh_ae2",
    "Sinh_aee",
    "Sinh_aee2",
    "Sinh_ai",
    "Sinh_ai2",
    "Sinh_al",
    "Sinh_au",
    "Sinh_au2",
    "Sinh_ba",
    "Sinh_bha",
    "Sinh_ca",
    "Sinh_cha",
    "Sinh_dda",
    "Sinh_ddha",
    "Sinh_dha",
    "Sinh_dhha",
    "Sinh_e",
    "Sinh_e2",
    "Sinh_ee",
    "Sinh_ee2",
    "Sinh_fa",
    "Sinh_ga",
    "Sinh_gha",
    "Sinh_h2",
    "Sinh_ha",
    "Sinh_i",
    "Sinh_i2",
    "Sinh_ii",
    "Sinh_ii2",
    "Sinh_ja",
    "Sinh_jha",
    "Sinh_jnya",
    "Sinh_ka",
    "Sinh_kha",
    "Sinh_kunddaliya",
    "Sinh_la",
    "Sinh_lla",
    "Sinh_lu",
    "Sinh_lu2",
    "Sinh_luu",
    "Sinh_luu2",
    "Sinh_ma",
    "Sinh_mba",
    "Sinh_na",
    "Sinh_ndda",
    "Sinh_ndha",
    "Sinh_ng",
    "Sinh_ng2",
    "Sinh_nga",
    "Sinh_nja",
    "Sinh_nna",
    "Sinh_nya",
    "Sinh_o",
    "Sinh_o2",
    "Sinh_oo",
    "Sinh_oo2",
    "Sinh_pa",
    "Sinh_pha",
    "Sinh_ra",
    "Sinh_ri",
    "Sinh_rii",
    "Sinh_ru2",
    "Sinh_ruu2",
    "Sinh_sa",
    "Sinh_sha",
    "Sinh_ssha",
    "Sinh_tha",
    "Sinh_thha",
    "Sinh_tta",
    "Sinh_ttha",
    "Sinh_u",
    "Sinh_u2",
    "Sinh_uu",
    "Sinh_uu2",
    "Sinh_va",
    "Sinh_ya",
    "sixsubscript",
    "sixsuperior",
    "slash",
    "SlowKeys_Enable",
    "soliddiamond",
    "space",
    "squareroot",
    "ssharp",
    "sterling",
    "StickyKeys_Enable",
    "stricteq",
    "Super_L",
    "Super_R",
    "Sys_Req",
    "t",
    "T",
    "Tab",
    "tabovedot",
    "Tabovedot",
    "tcaron",
    "Tcaron",
    "tcedilla",
    "Tcedilla",
    "telephone",
    "telephonerecorder",
    "Terminate_Server",
    "Thai_baht",
    "Thai_bobaimai",
    "Thai_chochan",
    "Thai_chochang",
    "Thai_choching",
    "Thai_chochoe",
    "Thai_dochada",
    "Thai_dodek",
    "Thai_fofa",
    "Thai_fofan",
    "Thai_hohip",
    "Thai_honokhuk",
    "Thai_khokhai",
    "Thai_khokhon",
    "Thai_khokhuat",
    "Thai_khokhwai",
    "Thai_khorakhang",
    "Thai_kokai",
    "Thai_lakkhangyao",
    "Thai_lekchet",
    "Thai_lekha",
    "Thai_lekhok",
    "Thai_lekkao",
    "Thai_leknung",
    "Thai_lekpaet",
    "Thai_leksam",
    "Thai_leksi",
    "Thai_leksong",
    "Thai_leksun",
    "Thai_lochula",
    "Thai_loling",
    "Thai_lu",
    "Thai_maichattawa",
    "Thai_maiek",
    "Thai_maihanakat",
    "Thai_maihanakat_maitho",
    "Thai_maitaikhu",
    "Thai_maitho",
    "Thai_maitri",
    "Thai_maiyamok",
    "Thai_moma",
    "Thai_ngongu",
    "Thai_nikhahit",
    "Thai_nonen",
    "Thai_nonu",
    "Thai_oang",
    "Thai_paiyannoi",
    "Thai_phinthu",
    "Thai_phophan",
    "Thai_phophung",
    "Thai_phosamphao",
    "Thai_popla",
    "Thai_rorua",
    "Thai_ru",
    "Thai_saraa",
    "Thai_saraaa",
    "Thai_saraae",
    "Thai_saraaimaimalai",
    "Thai_saraaimaimuan",
    "Thai_saraam",
    "Thai_sarae",
    "Thai_sarai",
    "Thai_saraii",
    "Thai_sarao",
    "Thai_sarau",
    "Thai_saraue",
    "Thai_sarauee",
    "Thai_sarauu",
    "Thai_sorusi",
    "Thai_sosala",
    "Thai_soso",
    "Thai_sosua",
    "Thai_thanthakhat",
    "Thai_thonangmontho",
    "Thai_thophuthao",
    "Thai_thothahan",
    "Thai_thothan",
    "Thai_thothong",
    "Thai_thothung",
    "Thai_topatak",
    "Thai_totao",
    "Thai_wowaen",
    "Thai_yoyak",
    "Thai_yoying",
    "therefore",
    "thinspace",
    "thorn",
    "Thorn",
    "THORN",
    "threeeighths",
    "threefifths",
    "threequarters",
    "threesubscript",
    "threesuperior",
    "tintegral",
    "topintegral",
    "topleftparens",
    "topleftradical",
    "topleftsqbracket",
    "topleftsummation",
    "toprightparens",
    "toprightsqbracket",
    "toprightsummation",
    "topt",
    "topvertsummationconnector",
    "Touroku",
    "trademark",
    "trademarkincircle",
    "tslash",
    "Tslash",
    "twofifths",
    "twosubscript",
    "twosuperior",
    "twothirds",
    "u",
    "U",
    "uacute",
    "Uacute",
    "ubelowdot",
    "Ubelowdot",
    "ubreve",
    "Ubreve",
    "ucircumflex",
    "Ucircumflex",
    "udiaeresis",
    "Udiaeresis",
    "udoubleacute",
    "Udoubleacute",
    "ugrave",
    "Ugrave",
    "uhook",
    "Uhook",
    "uhorn",
    "Uhorn",
    "uhornacute",
    "Uhornacute",
    "uhornbelowdot",
    "Uhornbelowdot",
    "uhorngrave",
    "Uhorngrave",
    "uhornhook",
    "Uhornhook",
    "uhorntilde",
    "Uhorntilde",
    "Ukrainian_ghe_with_upturn",
    "Ukrainian_GHE_WITH_UPTURN",
    "Ukrainian_i",
    "Ukrainian_I",
    "Ukrainian_ie",
    "Ukrainian_IE",
    "Ukrainian_yi",
    "Ukrainian_YI",
    "Ukranian_i",
    "Ukranian_I",
    "Ukranian_je",
    "Ukranian_JE",
    "Ukranian_yi",
    "Ukranian_YI",
    "umacron",
    "Umacron",
    "underbar",
    "underscore",
    "Undo",
    "union",
    "uogonek",
    "Uogonek",
    "Up",
    "uparrow",
    "upcaret",
    "upleftcorner",
    "uprightcorner",
    "upshoe",
    "upstile",
    "uptack",
    "uring",
    "Uring",
    "utilde",
    "Utilde",
    "v",
    "V",
    "variation",
    "vertbar",
    "vertconnector",
    "voicedsound",
    "VoidSymbol",
    "vt",
    "w",
    "W",
    "wacute",
    "Wacute",
    "wcircumflex",
    "Wcircumflex",
    "wdiaeresis",
    "Wdiaeresis",
    "wgrave",
    "Wgrave",
    "WonSign",
    "x",
    "X",
    "xabovedot",
    "Xabovedot",
    "XF86AddFavorite",
    "XF86ApplicationLeft",
    "XF86ApplicationRight",
    "XF86AudioCycleTrack",
    "XF86AudioForward",
    "XF86AudioLowerVolume",
    "XF86AudioMedia",
    "XF86AudioMicMute",
    "XF86AudioMute",
    "XF86AudioNext",
    "XF86AudioPause",
    "XF86AudioPlay",
    "XF86AudioPreset",
    "XF86AudioPrev",
    "XF86AudioRaiseVolume",
    "XF86AudioRandomPlay",
    "XF86AudioRecord",
    "XF86AudioRepeat",
    "XF86AudioRewind",
    "XF86AudioStop",
    "XF86Away",
    "XF86Back",
    "XF86BackForward",
    "XF86Battery",
    "XF86Blue",
    "XF86Bluetooth",
    "XF86Book",
    "XF86BrightnessAdjust",
    "XF86Calculater",
    "XF86Calculator",
    "XF86Calendar",
    "XF86CD",
    "XF86Clear",
    "XF86ClearGrab",
    "XF86Close",
    "XF86Community",
    "XF86ContrastAdjust",
    "XF86Copy",
    "XF86Cut",
    "XF86CycleAngle",
    "XF86Display",
    "XF86Documents",
    "XF86DOS",
    "XF86Eject",
    "XF86Excel",
    "XF86Explorer",
    "XF86Favorites",
    "XF86Finance",
    "XF86Forward",
    "XF86FrameBack",
    "XF86FrameForward",
    "XF86FullScreen",
    "XF86Game",
    "XF86Go",
    "XF86Green",
    "XF86Hibernate",
    "XF86History",
    "XF86HomePage",
    "XF86HotLinks",
    "XF86iTouch",
    "XF86KbdBrightnessDown",
    "XF86KbdBrightnessUp",
    "XF86KbdLightOnOff",
    "XF86Keyboard",
    "XF86Launch0",
    "XF86Launch1",
    "XF86Launch2",
    "XF86Launch3",
    "XF86Launch4",
    "XF86Launch5",
    "XF86Launch6",
    "XF86Launch7",
    "XF86Launch8",
    "XF86Launch9",
    "XF86LaunchA",
    "XF86LaunchB",
    "XF86LaunchC",
    "XF86LaunchD",
    "XF86LaunchE",
    "XF86LaunchF",
    "XF86LightBulb",
    "XF86LogGrabInfo",
    "XF86LogOff",
    "XF86LogWindowTree",
    "XF86Mail",
    "XF86MailForward",
    "XF86Market",
    "XF86Meeting",
    "XF86Memo",
    "XF86MenuKB",
    "XF86MenuPB",
    "XF86Messenger",
    "XF86ModeLock",
    "XF86MonBrightnessCycle",
    "XF86MonBrightnessDown",
    "XF86MonBrightnessUp",
    "XF86Music",
    "XF86MyComputer",
    "XF86MySites",
    "XF86New",
    "XF86News",
    "XF86Next_VMode",
    "XF86OfficeHome",
    "XF86Open",
    "XF86OpenURL",
    "XF86Option",
    "XF86Paste",
    "XF86Phone",
    "XF86Pictures",
    "XF86PowerDown",
    "XF86PowerOff",
    "XF86Prev_VMode",
    "XF86Q",
    "XF86Red",
    "XF86Refresh",
    "XF86Reload",
    "XF86Reply",
    "XF86RFKill",
    "XF86RockerDown",
    "XF86RockerEnter",
    "XF86RockerUp",
    "XF86RotateWindows",
    "XF86RotationKB",
    "XF86RotationLockToggle",
    "XF86RotationPB",
    "XF86Save",
    "XF86ScreenSaver",
    "XF86ScrollClick",
    "XF86ScrollDown",
    "XF86ScrollUp",
    "XF86Search",
    "XF86Select",
    "XF86Send",
    "XF86Shop",
    "XF86Sleep",
    "XF86Spell",
    "XF86SplitScreen",
    "XF86Standby",
    "XF86Start",
    "XF86Stop",
    "XF86Subtitle",
    "XF86Support",
    "XF86Suspend",
    "XF86Switch_VT_1",
    "XF86Switch_VT_10",
    "XF86Switch_VT_11",
    "XF86Switch_VT_12",
    "XF86Switch_VT_2",
    "XF86Switch_VT_3",
    "XF86Switch_VT_4",
    "XF86Switch_VT_5",
    "XF86Switch_VT_6",
    "XF86Switch_VT_7",
    "XF86Switch_VT_8",
    "XF86Switch_VT_9",
    "XF86TaskPane",
    "XF86Terminal",
    "XF86Time",
    "XF86ToDoList",
    "XF86Tools",
    "XF86TopMenu",
    "XF86TouchpadOff",
    "XF86TouchpadOn",
    "XF86TouchpadToggle",
    "XF86Travel",
    "XF86Ungrab",
    "XF86User1KB",
    "XF86User2KB",
    "XF86UserPB",
    "XF86UWB",
    "XF86VendorHome",
    "XF86Video",
    "XF86View",
    "XF86WakeUp",
    "XF86WebCam",
    "XF86WheelButton",
    "XF86WLAN",
    "XF86Word",
    "XF86WWAN",
    "XF86WWW",
    "XF86Xfer",
    "XF86Yellow",
    "XF86ZoomIn",
    "XF86ZoomOut",
    "y",
    "Y",
    "yacute",
    "Yacute",
    "ybelowdot",
    "Ybelowdot",
    "ycircumflex",
    "Ycircumflex",
    "ydiaeresis",
    "Ydiaeresis",
    "yen",
    "ygrave",
    "Ygrave",
    "yhook",
    "Yhook",
    "ytilde",
    "Ytilde",
    "z",
    "Z",
    "zabovedot",
    "Zabovedot",
    "zacute",
    "Zacute",
    "zcaron",
    "Zcaron",
    "Zenkaku",
    "Zenkaku_Hankaku",
    "Zen_Koho",
    "zerosubscript",
    "zerosuperior",
    "zstroke",
    "Zstroke",
    "button1",
    "button2",
    "button3",
    "button4",
    "button5",
    "button6",
    "button7",
    "button8",
    "button9",
    "button10",
    "button11",
    "button12",
    "button13",
    "button14",
    "button15",
    "button16",
    "button17",
    "button18",
    "button19",
    "button20",
    "button21",
    "button22",
    "button23",
    "button24",
}



"""Classes for managing sections and metadata in config files.

These classes parse sections and metadata found in the comments of sxhkdrc
files, in which different styles of formatting are handled specially.

As for sections, they may be "simple" in which there are no subsections, or
represent a "stack" in which there *are* subsections, which require that they
are completely enclosed by their parent sections.  There may also be *no*
sections below the root, in which no comments create new sections.

With regard to metadata, there may only be descriptions on single lines
("simple") or they may be key-value pairs across multiple lines ("key-value").

Many of them will also take regular expressions to configure the recognition of
sections and metadata.

Section classes:
    SectionHandler: abstract base class.
    RootSectionHandler: no new sections and places all keybinds in the root section.
    SimpleSectionHandler: flat sections.
    StackSectionHandler: subsections.

Metadata classes:
    MetadataParser: abstract base class.
    NullMetadataParser: no-op for all operations.
    SimpleDescriptionParser: description line immediately above the keybind.
    KeyValueMetadataParser: key-value pairs above the keybind.
"""
from __future__ import annotations

import re
from abc import ABC, abstractmethod, abstractproperty
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Pattern,
    Tuple,
    Union,
)

from .errors import MetadataParserError, SectionEOFError, SectionPushError
from .parser import Keybind, SpanTree, expand_sequences

__all__ = [
    # General.
    "SectionTreeNode",
    "find_enclosing_section",
    # Section handlers.
    "SectionHandler",
    "RootSectionHandler",
    "SimpleSectionHandler",
    "StackSectionHandler",
    # Metadata parsers.
    "MetadataParser",
    "NullMetadataParser",
    "SimpleDescriptionParser",
    "KeyValueMetadataParser",
]


# TODO: should sections have metadata too?
# Invariants:
#   - each section completely encloses all subsections recursively
@dataclass
class SectionTreeNode:
    """Node representing a section in the sxhkdrc.

    Note that only the root section will have `name` be `None`.
    """

    name: Optional[str]
    start: int
    end: Optional[int]
    children: List[SectionTreeNode]
    keybind_children: List[Keybind]

    def __init__(self, name: Optional[str], start: int, end: Optional[int]):
        self.name = name
        self.start = start
        # initially None when created
        self.end = end
        self.children = []
        self.keybind_children = []

    def add_child(
        self, name: str, start: int, end: Optional[int]
    ) -> SectionTreeNode:
        """Add a subsection with the given name and line range."""
        child = SectionTreeNode(name, start, end)
        self.children.append(child)
        return child

    def add_keybind(self, keybind: Keybind) -> None:
        """Add `keybind` to this section as a direct child."""
        if keybind.line is not None:
            assert keybind.line >= self.start
            if self.end is not None:
                assert keybind.line <= self.end
        self.keybind_children.append(keybind)

    @staticmethod
    def _get_level_prefix(level: int) -> str:
        return f"{' ' * (level-1)}└{'─' * (level-1)}"

    @staticmethod
    def _default_keybind_child_callback(keybind: Keybind, level: int) -> None:
        msg = f"{SectionTreeNode._get_level_prefix(level)}"
        if "description" in keybind.metadata:
            desc = keybind.metadata["description"]
            # TODO: handle exceptions
            expanded = expand_sequences(desc)
            if isinstance(expanded, str):
                assert desc == expanded
                # No permutations, so just print it plainly.
                msg = f"{msg} {keybind.hotkey.raw!r} (line {keybind.line})"
                if "mode" in keybind.metadata:
                    msg = f"{msg} (mode: {keybind.metadata['mode']})"
                msg = f"{msg}: {desc}"
                print(msg)
                return
            else:
                assert isinstance(expanded, SpanTree)
                # Try to print each permutation separately.
                permutations = expanded.generate_permutations()
                if len(permutations) == len(keybind.hotkey.permutations):
                    prefix = msg
                    for hotkey_perm, desc_perm in zip(
                        keybind.hotkey.permutations, permutations
                    ):
                        msg = f"{prefix} {str(hotkey_perm)!r} (line {keybind.line})"
                        if "mode" in keybind.metadata:
                            msg = f"{msg} (mode: {keybind.metadata['mode']})"
                        msg = f"{msg}: {desc_perm}"
                        print(msg)
                else:
                    if isinstance(keybind.hotkey.raw, str):
                        msg = f"{msg} {keybind.hotkey.raw!r} (line {keybind.line})"
                    else:
                        msg = f"{msg} {' '.join(keybind.hotkey.raw)!r} (line {keybind.line})"
                    if "mode" in keybind.metadata:
                        msg = f"{msg} (mode: {keybind.metadata['mode']})"
                    msg = f"{msg}: {desc}"
                    print(msg)
                return

        else:
            if isinstance(keybind.hotkey.raw, str):
                msg = f"{msg} {keybind.hotkey.raw!r} (line {keybind.line})"
            else:
                msg = f"{msg} {' '.join(keybind.hotkey.raw)!r} (line {keybind.line})"
            if "mode" in keybind.metadata:
                msg = f"{msg} (mode: {keybind.metadata['mode']})"
            print(msg)

    def _print_tree_rec(
        self,
        level: int,
        keybind_child_callback: Callable[[Keybind, int], None],
    ) -> None:
        assert level >= 0
        pos = (self.start, self.end)
        if level == 0:
            print(f"{self.name} {pos}")
        else:
            print(f"{' ' * (level-1)}└{'─' * (level-1)} {self.name} {pos}")
        # Print this section's keybinds under this node.
        for keybind in self.keybind_children:
            keybind_child_callback(keybind, level + 1)
        # And now the descendants.
        for child in self.children:
            child._print_tree_rec(level + 1, keybind_child_callback)

    def print_tree(
        self,
        keybind_child_callback: Optional[
            Callable[[Keybind, int], None]
        ] = None,
    ) -> None:
        """Print the section tree rooted at this node, with keybinds."""
        if not keybind_child_callback:
            keybind_child_callback = (
                SectionTreeNode._default_keybind_child_callback
            )
        self._print_tree_rec(0, keybind_child_callback)

    @classmethod
    def build_root(cls) -> SectionTreeNode:
        """Return a new root node."""
        return cls(None, 1, None)


# recursive and works because:
#   - base case: empty list
#   - few cases above base: child_gaps is a flat list of (node,gap) pairs
def _find_enclosing_section_rec(
    node: SectionTreeNode, keybind: Keybind
) -> List[Tuple[SectionTreeNode, int]]:
    assert node.start is not None, node
    assert node.end is not None, node
    assert keybind.line is not None, keybind
    if not (node.start <= keybind.line <= node.end):
        return []
    gap = abs(keybind.line - node.start) + abs(node.end - keybind.line)
    child_gaps = []
    for child in node.children:
        child_gaps.extend(_find_enclosing_section_rec(child, keybind))
    return [(node, gap)] + child_gaps


def find_enclosing_section(
    node: SectionTreeNode, keybind: Keybind
) -> Optional[SectionTreeNode]:
    """Find smallest enclosing section that is a descendant of `node` for `keybind`.

    Assumes that `keybind` has a non-None line number.
    """
    assert keybind.line is not None, keybind
    gaps = []
    for child in node.children:
        gaps.extend(_find_enclosing_section_rec(child, keybind))
    # sort by gap size
    gaps.sort(key=lambda x: x[1])
    if gaps:
        # smallest gap -> smallest section that contains the keybind
        section, gap = gaps[0]
        return section
    else:
        return None


class SectionHandler(ABC):
    """Abstract base class for managing sxhkdrc sections.

    Abstract methods/properties:
        - reset
        - clone_config
        - push
        - root
        - current_section
    """

    @abstractmethod
    def reset(self) -> None:
        """Reset the section handler."""
        raise NotImplementedError

    @abstractmethod
    def clone_config(self) -> SectionHandler:
        """Create an empty instance with the same configuration."""
        raise NotImplementedError

    @abstractmethod
    def push(self, text: str, line: int) -> bool:
        """Return whether `text` can be parsed as a section header or footer.

        This should be used for delimiting sections and metadata comments.
        """
        raise NotImplementedError

    def push_eof(self, last_line: int) -> None:
        """Do clean-up actions after the input ends.

        This should be called by overriding subclasses so that the root section
        gets its `end` attribute defined.
        """
        self.root.end = last_line

    @abstractproperty
    def root(self) -> SectionTreeNode:
        """Return the root section."""
        raise NotImplementedError

    @abstractproperty
    def current_section(self) -> SectionTreeNode:
        """Return the current section."""
        raise NotImplementedError


@dataclass
class RootSectionHandler(SectionHandler):
    """Handler for a single-section sxhkdrc, where all keybinds are children of the root."""

    _section: SectionTreeNode

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """Reset the section handler."""
        self._section = SectionTreeNode.build_root()

    def clone_config(self) -> RootSectionHandler:
        """Create an empty instance with the same configuration."""
        return RootSectionHandler()

    def push(self, text: str, line: int) -> bool:
        """Reject the input.

        No new sections will be defined.
        """
        return False

    @property
    def root(self) -> SectionTreeNode:
        """Return the root section (which is the only node)."""
        return self._section

    @property
    def current_section(self) -> SectionTreeNode:
        """Return the current section (which is always the root)."""
        return self._section


@dataclass
class SimpleSectionHandler(SectionHandler):
    """Handler for sections one level under the root section.

    Each new section is a direct child of the root and each such section ends
    when a new section begins or EOF.

    Instance variables:
        section_header_re: the pattern to match and parse out section headers.
        sections: the sections of the config in the order they were defined.
    """

    section_header_re: Pattern[str]
    sections: List[SectionTreeNode]
    _root: SectionTreeNode = field(repr=False)

    def __init__(self, section_header_re: Union[str, re.Pattern[str]]):
        """Create an instance with a regex for section headers.

        `section_header_re` must have a named group 'name'.
        """
        if isinstance(section_header_re, re.Pattern):
            self.section_header_re = section_header_re
        else:
            self.section_header_re = re.compile(section_header_re)
        if "name" not in self.section_header_re.groupindex:
            raise ValueError(
                "section header regex must have the named group 'name'"
            )
        self.reset()

    def reset(self) -> None:
        """Reset the section handler."""
        self._root = SectionTreeNode.build_root()
        self.sections = [self._root]

    def clone_config(self) -> SimpleSectionHandler:
        """Create an empty instance with the same configuration."""
        return SimpleSectionHandler(self.section_header_re)

    def push(self, text: str, line: int) -> bool:
        """Return whether `text` can be parsed as a section header.

        If `section_header_re` matches `text`, it is a section header: a new
        section is added and the previous section ended.
        """
        m = self.section_header_re.search(text)
        if m:
            # starting a new section ends the previous one
            if self.sections[-1] is not self._root:
                self.sections[-1].end = line - 1
            # self.sections[self._curr_section_name] = (self._curr_section_start, end)
            self.sections.append(
                self._root.add_child(
                    m.group("name"),
                    line,
                    None,
                )
            )
            return True
        else:
            return False

    @property
    def root(self) -> SectionTreeNode:
        """Return the root section."""
        return self._root

    @property
    def current_section(self) -> SectionTreeNode:
        """Return the latest section."""
        return self.sections[-1]


@dataclass
class StackSectionHandler(SectionHandler):
    """Handler for recursive sections and subsections.

    This uses a stack to keep track of sections, and requires that inner
    sections be merged before their parent sections.

    Instance variables:
        section_header_re: the pattern to match and parse out section headers.
        section_footer_re: the pattern to match section footers.
    """

    section_header_re: Pattern[str]
    section_footer_re: Pattern[str]
    _section_tree: SectionTreeNode = field(repr=False)
    _section_stack: List[SectionTreeNode] = field(repr=False)

    def __init__(
        self,
        section_header_re: Union[str, re.Pattern[str]],
        section_footer_re: Union[str, re.Pattern[str]],
    ):
        """Create an instance with regexes for section headers and footers.

        `section_header_re` must have a named group 'name'.
        `section_footer_re` doesn't need any named groups.
        """
        if isinstance(section_header_re, re.Pattern):
            self.section_header_re = section_header_re
        else:
            self.section_header_re = re.compile(section_header_re)
        if isinstance(section_footer_re, re.Pattern):
            self.section_footer_re = section_footer_re
        else:
            self.section_footer_re = re.compile(section_footer_re)
        if "name" not in self.section_header_re.groupindex:
            raise ValueError(
                "section header regex must have the named group 'name'"
            )
        self.reset()

    def reset(self) -> None:
        """Reset the section handler."""
        self._section_tree = SectionTreeNode.build_root()
        self._section_stack = [self._section_tree]

    def clone_config(self) -> StackSectionHandler:
        """Create an empty instance with the same configuration."""
        return StackSectionHandler(
            self.section_header_re, self.section_footer_re
        )

    def push(self, text: str, line: int) -> bool:
        """Return whether `text` can be parsed as a section header or footer.

        If `section_header_re` matches `text`, it is a section header: a new
        section is created and added to the stack.

        If `section_footer_re` matches `text`, it is a section footer: the
        current section's ending is defined and it is popped from the stack.

        NOTE: ending a section before any have been defined raises SectionPushError.
        """
        m = self.section_header_re.search(text)
        if m:
            name = m.group("name")
            node = self._section_stack[-1].add_child(name, line, None)
            self._section_stack.append(node)
            return True
        else:
            m = self.section_footer_re.search(text)
            if m:
                # first *must* always be root
                if len(self._section_stack) == 1:
                    raise SectionPushError(
                        "Ended a section without opening one first",
                        line=line,
                    )
                node = self._section_stack.pop()
                node.end = line
                return True
            else:
                return False

    def push_eof(self, last_line: int) -> None:
        """Ensure no sections have been left unclosed.

        If any *have* been left open, SectionEOFError is raised.
        Otherwise, defines the ending line number for the root section.
        """
        if len(self._section_stack) > 1:
            raise SectionEOFError(
                f"Got EOF while reading section '{self._section_stack[-1].name}'",
                last_line=last_line,
                sections=self._section_stack[1:],
            )
        super().push_eof(last_line)

    @property
    def root(self) -> SectionTreeNode:
        """Return the root section."""
        return self._section_tree

    @property
    def current_section(self) -> SectionTreeNode:
        """Return the section at the top of the stack."""
        return self._section_stack[-1]


class MetadataParser(ABC):
    """Abstract base class for parsing metadata comments for a keybind.

    Abstract methods:
        - parse
    """

    @abstractmethod
    def parse(self, lines: Iterable[str], start_line: int) -> Dict[str, Any]:
        """Parse metadata from the comments immediately preceding a keybind."""
        raise NotImplementedError


@dataclass
class NullMetadataParser(MetadataParser):
    """Parser that always returns an empty mapping when parsing."""

    def parse(self, lines: Iterable[str], start_line: int) -> Dict[str, Any]:
        """Return an empty dict."""
        return {}


@dataclass
class SimpleDescriptionParser(MetadataParser):
    """Parser for the description line immediately above a keybind.

    Instance variables:
        description_re: the pattern to parse out the description from the comment.
    """

    description_re: Pattern[str]

    def __init__(self, description_re: str):
        """Create an instance with a regex for description lines.

        `description_re` must have a named group 'description'.
        """
        self.description_re = re.compile(description_re)
        if "description" not in self.description_re.groupindex:
            raise ValueError(
                "description regex must have the named group 'description'"
            )

    def parse(self, lines: Iterable[str], start_line: int) -> Dict[str, Any]:
        """Parse a description from the last element of `lines`.

        If the regex match with `description_re` succeeds, a dict with the
        key-value pair "description" and the 'description' match group is
        returned.  Otherwise, an empty dict is returned instead.
        """
        comments = list(lines)
        assert len(comments) > 0
        maybe_description = comments[-1]
        m = self.description_re.search(maybe_description)
        if m:
            return {"description": m.group("description")}
        else:
            return {}


@dataclass
class KeyValueMetadataParser(MetadataParser):
    """Parser for key-value pairs in metadata.

    Instance variables:
        pair_re: the pattern to parse out key-value pairs.
        empty_re: the pattern to match empty metadata lines.
    """

    pair_re: Pattern[str]
    empty_re: Pattern[str]  # part of description but no pair

    def __init__(self, pair_re: str, empty_re: str):
        """Create an instance with regexes for pairs and empty metadata lines.

        `pair_re` must have the named groups 'key' and 'value'.
        `empty_re` doesn't need any named groups.
        """
        self.pair_re = re.compile(pair_re)
        if (
            "key" not in self.pair_re.groupindex
            or "value" not in self.pair_re.groupindex
        ):
            raise ValueError(
                "pair regex must have named groups 'key' and 'value'"
            )
        self.empty_re = re.compile(empty_re)

    def parse(self, lines: Iterable[str], start_line: int) -> Dict[str, Any]:
        """Parse metadata comments for key-value pairs.

        They are parsed in the reverse order of `lines` so that the contiguous
        block of metadata comments immediately preceding a keybind isn't
        interrupted by non-metadata comments.

        `empty_re` is used to match comments that are part of that contiguous
        block, but don't define key-value pairs.  Useful for "blank" lines
        between key-value pairs or for comments about them in a separate line.

        NOTE: duplicate keys cause MetadataParserError to be raised.
        """
        comments = list(lines)
        # Each dict value stores the line number of the pair that is latest in
        # the file, so that errors on duplicates show the actual duplicate
        # rather than the first occurrence.
        metadata: Dict[str, Tuple[int, Any]] = {}
        i = len(comments) - 1
        while i >= 0:
            comment = comments[i]
            m = self.pair_re.search(comment)
            if m:
                key = m.group("key")
                value = m.group("value")
                if key in metadata:
                    dup_line, _ = metadata[key]
                    raise MetadataParserError(
                        f"Duplicate key '{key}'",
                        key=key,
                        value=value,
                        line=dup_line,
                    )
                metadata[key] = (start_line + i, value)
                i -= 1
            elif self.empty_re.search(comment):
                # metadata continues
                i -= 1
                continue
            else:
                break
        return {key: val for key, (_, val) in metadata.items()}



"""Library for parsing Simple X Hotkey Daemon (sxhkd) keybinds.

You should start with the `parser` module.

Re-exports every member in all modules directly part of this package.  The
`cli` subpackage as well as modules under it need to be imported explicitly.
"""
# mypy: implicit-reexport
from ._package import *
from .errors import *
from .keysyms import *
from .metadata import *
from .parser import *
from .util import *



"""Exceptions for the library.

SXHKDParserError is the ancestor for almost all of the exceptions in the
library.  Some functions and methods raise ValueError, but they are few and
small in scope.
"""
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
)

if TYPE_CHECKING:
    from .metadata import SectionTreeNode
    from .parser import (
        HotkeyPermutation,
        HotkeyToken,
        HotkeyTreeLeafNode,
        _HotkeyParseMode,
    )

    TransitionTable = Dict[
        str, Tuple[_HotkeyParseMode, Callable[[HotkeyToken], None]]
    ]

__all__ = [
    "SXHKDParserError",
    # ---
    "ConfigReadError",
    "MissingHotkeyError",
    # ---
    "KeybindError",
    "InconsistentKeybindCasesError",
    # ---
    "SequenceParseError",
    # ---
    "HotkeyError",
    "HotkeyTokenizeError",
    "HotkeyParseError",
    "UnexpectedTokenError",
    "NonTerminalStateExitError",
    "InconsistentNoabortError",
    "DuplicateModifierError",
    "DuplicateChordPermutationError",
    "ConflictingChainPrefixError",
    "PossiblyInvalidKeysyms",
    # ---
    "SectionHandlerError",
    "SectionPushError",
    "SectionEOFError",
    # ---
    "MetadataParserError",
]


class SXHKDParserError(Exception):
    """Ancestor for almost all of the exceptions in the library."""

    def __init__(
        self,
        message: str,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ):
        super().__init__(message)
        self.message = message
        self.line = line
        self.column = column

    def __str__(self) -> str:
        if self.line is not None:
            if self.column is None:
                return f"{self.line}: {self.message}"
            else:
                return f"{self.line}:{self.column}: {self.message}"
        elif self.column is not None:
            return f"{self.message} at column {self.column}"
        else:
            return self.message


class ConfigReadError(SXHKDParserError):
    """Base class for errors related to reading sxhkdrc files."""

    pass


class MissingHotkeyError(ConfigReadError):
    """Command text was read without an accompanying hotkey."""

    pass


class KeybindError(SXHKDParserError):
    """Base class for errors related to `Keybind` objects."""

    pass


class InconsistentKeybindCasesError(KeybindError):
    """A keybind had an inconsistent number of cases."""

    def __init__(
        self,
        message: str,
        hotkey_cases: int,
        command_cases: int,
        line: Optional[int] = None,
    ):
        super().__init__(message=message, line=line)
        self.hotkey_cases = hotkey_cases
        self.command_cases = command_cases


class SequenceParseError(SXHKDParserError):
    """A sequence of the form {s1,s2,...,sn} failed to parse."""

    def __init__(
        self,
        message: str,
        text: str,
        line: Optional[int] = None,
        column: Optional[int] = None,
    ):
        super().__init__(message=message, line=line, column=column)
        self.text = text
        self.line = line


class HotkeyError(SXHKDParserError):
    """Base class for errors related to `Hotkey` objects."""

    pass


class HotkeyTokenizeError(HotkeyError):
    """An invalid character was passed to the tokenizer."""

    def __init__(
        self, message: str, hotkey: str, value: str, line: Optional[int] = None
    ):
        super().__init__(message=message, line=line)
        self.hotkey = hotkey
        self.value = value


class HotkeyParseError(HotkeyError):
    """Base class for issues related to parsing hotkeys into Hotkey objects."""

    pass


class UnexpectedTokenError(HotkeyParseError):
    """A token was in the wrong position."""

    def __init__(
        self,
        message: str,
        token: HotkeyToken,
        mode: _HotkeyParseMode,
        transitions: TransitionTable,
        tokens: List[HotkeyToken],
    ):
        super().__init__(message=message)
        self.token = token
        self.mode = mode
        self.transitions = transitions
        self.tokens = tokens


class NonTerminalStateExitError(HotkeyParseError):
    """The input ended on a non-terminal parser state."""

    def __init__(self, message: str, mode: _HotkeyParseMode):
        super().__init__(message=message)
        self.mode = mode


class InconsistentNoabortError(HotkeyParseError):
    """The colon character was used in different places along the permutations of a hotkey."""

    def __init__(
        self,
        message: str,
        perm1: HotkeyPermutation,
        perm1_index: int,
        perm2: HotkeyPermutation,
        perm2_index: int,
        line: Optional[int] = None,
    ):
        super().__init__(message=message, line=line)
        self.perm1 = perm1
        self.perm1_index = perm1_index
        self.perm2 = perm2
        self.perm2_index = perm2_index


class DuplicateModifierError(HotkeyParseError):
    """A modifier was repeated in the same chord."""

    def __init__(self, message: str, modifier: str):
        super().__init__(message=message)
        self.modifier = modifier


class DuplicateChordPermutationError(HotkeyError):
    """A chord permutation was a duplicate of another."""

    def __init__(
        self,
        message: str,
        dup_perm: HotkeyPermutation,
        perm1_index: int,
        perm2_index: int,
        line: Optional[int] = None,
    ):
        super().__init__(message=message, line=line)
        self.dup_perm = dup_perm
        self.perm1_index = perm1_index
        self.perm2_index = perm2_index


class ConflictingChainPrefixError(HotkeyError):
    """An entire permutation of a hotkey was a prefix of another."""

    def __init__(
        self,
        message: str,
        chain_prefix: HotkeyTreeLeafNode,
        conflicts: List[HotkeyTreeLeafNode],
        line: Optional[int] = None,
    ):
        super().__init__(message=message, line=line)
        self.chain_prefix = chain_prefix
        self.conflicts = conflicts


class PossiblyInvalidKeysyms(HotkeyError):
    """Possibly invalid keysyms were found in a hotkey.

    Based on the keysyms found in X11 include files `keysymdef.h` and
    `XF86keysym.h` from Debian Bullseye's `x11proto-dev` package.

    There may be non-standard keysyms that people may want to use, so this
    shouldn't be fatal by default.
    """

    def __init__(
        self,
        message: str,
        keysyms: Set[str],
        line: Optional[int] = None,
    ):
        super().__init__(message=message, line=line)
        self.keysyms = keysyms


class SectionHandlerError(SXHKDParserError):
    """Base class for issues related to the management of sections."""

    pass


class SectionPushError(SectionHandlerError):
    """Miscellaneous errors while pushing a potential section."""

    def __init__(self, message: str, line: int):
        super().__init__(message=message, line=line)


class SectionEOFError(SectionHandlerError):
    """Miscellaneous errors after receiving EOF."""

    def __init__(
        self, message: str, last_line: int, sections: List[SectionTreeNode]
    ):
        super().__init__(message=message, line=last_line)
        self.message = message
        self.last_line = last_line
        self.sections = sections


class MetadataParserError(SXHKDParserError):
    """Miscellaneous errors while instance was parsing comments for metadata."""

    def __init__(self, message: str, key: str, value: Any, line: int):
        super().__init__(message=message, line=line)
        self.key = key
        self.value = value
